# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import json
from enum import Enum as enum_t
from pathlib import Path as path_t
from typing import Any, Callable, Dict, Optional, Tuple, Union

try:
    import networkx as grph
except ModuleNotFoundError:
    grph = None
try:
    import numpy as nmpy
except ModuleNotFoundError:
    nmpy = None

from cell_tracking_BC.in_out.text.logger import LOGGER


builders_h = Dict[str, Callable[[Union[str, dict]], Any]]
description_h = Tuple[str, Any]
# /!\ When a module is not found, using bytes, the first type tested while JSONing, as the main module type is a safe
# way to "disable" it.
if grph is None:
    graph_t = bytes
else:
    graph_t = grph.Graph
if nmpy is None:
    array_t = bytes
else:
    array_t = nmpy.ndarray
object_h = Union[bytes, slice, str, list, tuple, dict, path_t, array_t, graph_t, Any]


_ERROR_MARKER = "!"  # Must be a character forbidden in type names
_TYPES_SEPARATOR = "@"  # Must be a character forbidden in type names


def JsonStringOf(instance: Any, /, *, true_type: Union[Any, type] = None) -> str:
    """"""
    if not ((true_type is None) or (type(true_type) is type)):
        true_type = type(true_type)

    return json.dumps(_JsonDescriptionOf(instance, 0, true_type=true_type))


def _JsonDescriptionOf(
    instance: Any, calling_level: int, /, *, true_type: type = None
) -> description_h:
    """"""
    # Test for high-level classes first since they can also be subclasses of standard classes below, but only if not
    # called from the instance.AsJsonString method, which would create infinite recursion.
    if (calling_level > 0) and hasattr(instance, "AsJsonString"):
        return _JsonDescriptionOf(
            instance.AsJsonString(),
            calling_level + 1,
            true_type=_AutomaticTrueType(instance, true_type),
        )

    if dtcl.is_dataclass(instance):
        # Do not use dtcl.asdict(self) since it recurses into dataclass instances which, if they extend a "container"
        # class like list or dict, will lose the contents.
        as_dict = {
            _fld.name: getattr(instance, _fld.name) for _fld in dtcl.fields(instance)
        }
        true_type = _AutomaticTrueType(instance, true_type)

        return _JsonDescriptionOf(as_dict, calling_level + 1, true_type=true_type)

    error = ""

    # /!\ Must be the first type to be tested (see unfoundable modules above)
    if isinstance(instance, bytes):
        base_type, jsonable = "bytes", instance.hex()
    elif isinstance(instance, enum_t):
        true_type = _AutomaticTrueType(instance, true_type)
        base_type, jsonable = "enum_Enum", _JsonDescriptionOf(
            instance.value, calling_level + 1
        )
    elif isinstance(instance, slice):
        base_type, jsonable = "slice", (instance.start, instance.stop, instance.step)
    elif isinstance(instance, (list, tuple)):
        base_type, jsonable = (
            type(instance).__name__,
            [_JsonDescriptionOf(_elm, calling_level + 1) for _elm in instance],
        )
    elif isinstance(instance, dict):
        # json does not accept non-str dictionary keys, hence the json.dumps
        base_type, jsonable = (
            "dict",
            {
                json.dumps(
                    _JsonDescriptionOf(_key, calling_level + 1)
                ): _JsonDescriptionOf(_vle, calling_level + 1)
                for _key, _vle in instance.items()
            },
        )
    elif isinstance(instance, path_t):
        base_type, jsonable = "pathlib_Path", str(instance)
    elif isinstance(instance, array_t):
        base_type, jsonable = "numpy_ndarray", (instance.tolist(), instance.dtype.name)
    elif isinstance(instance, graph_t):
        edges = grph.to_dict_of_dicts(instance)
        # /!\ Node attributes are added to the edges dictionary! This must be taken into account when deJSONing. Note
        # that several attempts to avoid this have been made, including one relying on repr(node), which is based on
        # hash(node). Since the hash function gives different results across Python sessions, this could not work.
        for node, attributes in instance.nodes(data=True):
            edges[node] = (attributes, edges[node])
        base_type, jsonable = (
            "networkx_Graph",
            (
                _JsonDescriptionOf(edges, calling_level + 1),
                type(instance).__name__,
            ),
        )
    else:
        base_type = type(instance).__name__
        try:
            _ = json.dumps(instance)
            jsonable = instance
        except TypeError:
            jsonable = None
            error = _ERROR_MARKER
            LOGGER.warn(f"{base_type}: UnJSONable type. Using None.")

    if true_type is None:
        true_type = ""
    else:
        true_type = true_type.__name__

    return f"{base_type}{_TYPES_SEPARATOR}{true_type}{error}", jsonable


def ObjectFromJsonString(jsoned: str, /, *, builders: builders_h = None) -> object_h:
    """"""
    return _ObjectFromJsonDescription(json.loads(jsoned), builders=builders)


def _ObjectFromJsonDescription(
    description: description_h,
    /,
    *,
    builders: builders_h = None,
) -> object_h:
    """"""
    types, instance = description
    base_type, true_type = types.split(_TYPES_SEPARATOR)
    if true_type.endswith(_ERROR_MARKER):
        true_type = true_type[: -_ERROR_MARKER.__len__()]
        error = True
    else:
        error = False
    if true_type.__len__() == 0:
        true_type = None
    if builders is None:
        builders = {}

    if error:
        LOGGER.warn(
            f"{base_type}{_TYPES_SEPARATOR}{true_type}: UnJSONable type. Returning None."
        )
        return None

    if base_type == "bytes":
        output = bytes.fromhex(instance)
    elif base_type == "enum_Enum":
        output = _ObjectFromJsonDescription(instance, builders=builders)
    elif base_type == "slice":
        output = slice(*instance)
    elif base_type in ("list", "tuple"):
        iterator = (
            _ObjectFromJsonDescription(_elm, builders=builders) for _elm in instance
        )
        if base_type == "list":
            output = list(iterator)
        else:
            output = tuple(iterator)
    elif base_type == "dict":
        output = {
            _ObjectFromJsonDescription(
                json.loads(_key), builders=builders
            ): _ObjectFromJsonDescription(_vle, builders=builders)
            for _key, _vle in instance.items()
        }
    elif base_type == "pathlib_Path":
        output = path_t(instance)
    elif base_type == "numpy_ndarray":
        data, dtype = instance
        output = nmpy.array(data, dtype=dtype)
    elif base_type == "networkx_Graph":
        edges_w_attributes, graph_type = instance
        edges_w_attributes = _ObjectFromJsonDescription(edges_w_attributes, builders=builders)
        graph_type = getattr(grph, graph_type)

        attributes = {}
        edges = {}
        for node, (node_attributes, edge) in edges_w_attributes.items():
            attributes[node] = node_attributes
            edges[node] = edge

        output = grph.from_dict_of_dicts(edges, create_using=graph_type)
        grph.set_node_attributes(output, attributes)
    else:
        output = instance

    if true_type in builders:
        output = builders[true_type](output)
    elif true_type is not None:
        output = None
        LOGGER.warn(f"{true_type}: Type without decoder. Returning None.")

    return output


def _AutomaticTrueType(instance: Any, true_type: Optional[type], /) -> type:
    """"""
    if true_type is None:
        return type(instance)

    # This was added to support passing a true type of an object built from a subclass instance. Usage example:
    # decomposition of an instance of a class with multiple inheritance into its components built from the instance
    # itself.
    if issubclass(type(instance), true_type):
        return true_type

    raise ValueError(
        f'{true_type.__name__}: Invalid true type specification for type "{type(instance).__name__}". Expected: None.'
    )



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import cell_tracking_BC.in_out.file.jsoner as jsnr
from cell_tracking_BC.type.cell import cell_t, state_e
from cell_tracking_BC.type.track.structured import (
    cells_w_optional_time_points_h,
    per_single_track_cells_h,
    structured_track_t,
)


# TODO: check if grph.descendants and grph.ancestors could be used in place of more complex code here
# TODO: cell iteration methods should have a geometric_mode: bool = False parameter to allow cell-state-independent
#     iterations


class single_track_t(structured_track_t, List[cell_t]):
    label: int = None
    affinities: Tuple[float, ...] = None
    _geometric_root_time_point: int = None

    def __init__(self, *args, **kwargs) -> None:
        """"""
        list.__init__(self, *args, **kwargs)
        # Place after list init to get access to self[...]
        structured_track_t.__init__(
            self,
            root=self[0],
            leaves=(self[-1],),
        )
        self._dividing_marked = True

    @classmethod
    def NewFromOrderedCells(
        cls,
        cells: Sequence[cell_t],
        affinities: Sequence[float],
        root_time_point: int,
        label: Optional[int],
        /,
    ) -> single_track_t:
        """
        This must be the only place where direct instantiation is allowed. Anywhere else, instantiation must be
        performed with this class method.

        label: Can be None only to accommodate the creation of branches as single tracks
        """
        n_cells = cells.__len__()
        if (n_affinities := affinities.__len__()) != (n_expected := n_cells - 1):
            raise ValueError(
                f"{n_affinities}: Invalid number of affinities. Expected={n_expected}"
            )

        instance = cls(cells)
        instance.label = label
        instance.affinities = tuple(affinities)
        instance._geometric_root_time_point = root_time_point

        return instance

    @classmethod
    def NewFromJsonString(
        cls,
        jsoned: str,
        /,
    ) -> single_track_t:
        """"""
        # See NewFromJsonDescription for why builders do not count one for structured_track_t
        return jsnr.ObjectFromJsonString(
            jsoned,
            builders={
                cls.__name__: cls.NewFromJsonDescription,
                "cell_t": cell_t.NewFromJsonString,
            },
        )

    @classmethod
    def NewFromJsonDescription(
        cls,
        description,
        /,
    ) -> single_track_t:
        """"""
        structured, cells, attributes = description
        # See AsJsonString for why decoding happens here
        structured = structured_track_t.NewFromJsonString(structured)

        instance = cls.NewFromOrderedCells(
            cells,
            attributes["affinities"],
            attributes["_geometric_root_time_point"],
            attributes["label"],
        )

        # Do not use dtcl.asdict(self) since it recurses into dataclass instances which, if they extend a "container"
        # class like list or dict, will lose the contents.
        attributes = {
            _fld.name: getattr(structured, _fld.name)
            for _fld in dtcl.fields(structured)
        }
        for key, value in attributes.items():
            setattr(instance, key, value)

        return instance

    @property
    def cells(self) -> Union[Iterable, Iterator]:
        """"""
        return self

    # @property
    # def label(self) -> Optional[int]:
    #     """"""
    #     return self._label

    @property
    def labels(self) -> Sequence[Optional[int]]:
        """"""
        return (self.label,)  # Note: this is a tuple

    @property
    def pruned_labels(self) -> Sequence[int]:
        """"""
        if self.IsFullyPruned():
            return self.labels

        return ()

    @property
    def length(self) -> int:
        """
        Segment-wise, not node-wise
        """
        # Do not use list length as it does not account for pruned parts
        return self.lengths[0]

    @property
    def leaf(self) -> cell_t:
        """"""
        return self.leaves[0]

    @property
    def leaf_time_point(self) -> int:
        """"""
        return self.CellTimePoint(self.leaf)

    def CellTimePoint(self, cell: cell_t) -> int:
        """"""
        return self._geometric_root_time_point + self.index(cell)

    def CellSuccessors(
        self, cell: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[Sequence[cell_t]]:
        """"""
        if (not check) or (cell in self):
            where = self.index(cell)
            if where < self.__len__() - 1:
                successor = self[where + 1]
                if successor.state is state_e.pruned:
                    output = ()
                else:
                    output = (successor,)
            else:
                output = ()

            return output

        if tolerant_mode:
            return None

        raise ValueError(f"{cell}: Cell not in track")

    def CellDescendants(
        self,
        cell: cell_t,
        /,
        *,
        including_self: bool = True,
        geometric_mode: bool = False,
        check: bool = True,
        tolerant_mode: bool = False,
    ) -> Optional[Sequence[cell_t]]:
        """"""
        if (not check) or (cell in self):
            where = self.index(cell)
            if not including_self:
                where += 1

            if where > self.__len__() - 1:
                output = ()
            else:
                output = tuple(self[where:])
                if not geometric_mode:
                    pruned = tuple(_cll.state is state_e.pruned for _cll in output)
                    if any(pruned):
                        first_pruned = pruned.index(True)
                        output = output[:first_pruned]

            return output

        if tolerant_mode:
            return None

        raise ValueError(f"{cell}: Cell not in track")

    def ConfirmCellLineage(self, youngest: cell_t, oldest: cell_t, /) -> bool:
        """"""
        if (youngest in self) and (oldest in self):
            return self.index(youngest) <= self.index(oldest)

        raise ValueError(
            f"{youngest}/{oldest}: Cell track memberships={youngest in self}/{oldest in self}. "
            f"Expected=True/True"
        )

    def SplittingCells(
        self, /, *, with_time_point: bool = False
    ) -> Sequence[Union[cell_t, Tuple[cell_t, int]]]:
        """"""
        return ()

    def MarkDividingCells(
        self,
        before_deaths: bool,
        /,
        *,
        division_responses: Dict[int, Optional[Sequence[float]]] = None,
        lower_bound: float = None,
        CombinedResponse: Callable[[Iterable], float] = max,
    ) -> None:
        """"""
        pass

    @property
    def n_dividing_cells(self) -> int:
        """"""
        return 0

    def DividingCells(
        self,
        /,
        *,
        with_time_point: bool = False,
        per_single_track: bool = False,
    ) -> Union[cells_w_optional_time_points_h, per_single_track_cells_h]:
        """"""
        return ()

    @property
    def unpruned_piece(self) -> single_track_t:
        """"""
        root_idx = self.index(self.root)
        leaf_idx = self.index(self.leaf)

        output = self.__class__.NewFromOrderedCells(
            self[root_idx : (leaf_idx + 1)],
            self.affinities[root_idx:leaf_idx],
            self._geometric_root_time_point + root_idx,
            self.label,
        )
        output._dead_marked = self._dead_marked

        return output

    @property
    def segments_iterator(self) -> Iterator[Tuple[int, cell_t, cell_t, bool]]:
        """"""
        unpruned = self.unpruned_piece

        n_cells = unpruned.__len__()
        for c_idx in range(1, n_cells):
            time_point = self.root_time_point + c_idx - 1
            is_last = c_idx == n_cells - 1
            yield time_point, *unpruned[(c_idx - 1) : (c_idx + 1)], is_last

    def Pieces(self, /, **_) -> Sequence[single_track_t]:
        """"""
        return (self.unpruned_piece,)  # Note: this is a tuple

    def TrackLabelsContainingCell(
        self, cell: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[Sequence[int]]:
        """"""
        if (not check) or (cell in self):
            return self.labels

        if tolerant_mode:
            return None

        raise ValueError(f"{cell}: Cell not in track")

    def TrackLabelWithLeaf(
        self, leaf: cell_t, /, *, check: bool = True, tolerant_mode: bool = False, geometric_mode: bool = False,
    ) -> Optional[int]:
        """"""
        if geometric_mode:
            leaves = self.geometric_leaves
        else:
            leaves = self.leaves
        if (not check) or (leaf in leaves):
            return self.label

        if tolerant_mode:
            return None

        raise ValueError(f"{leaf}: Not a leaf cell")

    def TrackWithLeaf(
        self, leaf: cell_t, /, *, check: bool = True, tolerant_mode: bool = False,
        geometric_mode: bool = False,
    ) -> Optional[single_track_t]:
        """"""
        if geometric_mode:
            leaves = self.geometric_leaves
        else:
            leaves = self.leaves
        if (not check) or (leaf in leaves):
            if geometric_mode:
                return self
            else:
                return self.unpruned_piece

        if tolerant_mode:
            return None

        raise ValueError(f"{leaf}: Not a leaf cell")

    def AsSingleTrack(self) -> single_track_t:
        """"""
        return self.unpruned_piece

    def AsRowsColsTimes(
        self, /, *, with_labels: bool = False
    ) -> Union[
        Tuple[Tuple[float, ...], Tuple[float, ...], Tuple[int, ...]],
        Tuple[Tuple[float, ...], Tuple[float, ...], Tuple[int, ...], Tuple[int, ...]],
    ]:
        """"""
        unpruned = self.unpruned_piece
        n_cells = unpruned.__len__()

        rows, cols = tuple(zip(*(_cll.centroid.tolist() for _cll in unpruned)))
        times = tuple(range(self.root_time_point, self.root_time_point + n_cells))

        if with_labels:
            labels = tuple(_cll.label for _cll in unpruned)
            return rows, cols, times, labels

        return rows, cols, times

    def CellFeature(self, feature: str, /, *, geometric_mode: bool = False) -> Sequence[Any]:
        """
        Contrary to structured_track_t.features which are rooted at time point zero using Nones if needed, here the
        feature is returned for the unpruned part only.
        """
        if geometric_mode:
            return tuple(_cll.features[feature] for _cll in self)

        return tuple(_cll.features[feature] for _cll in self.unpruned_piece)

    def AsJsonString(self) -> str:
        """"""
        structured = structured_track_t.AsJsonString(self)
        attributes = {
            "label": self.label,
            "affinities": self.affinities,
            "_geometric_root_time_point": self._geometric_root_time_point,
        }

        # list(self): To hide this method, which would otherwise be infinitely recursively called
        return jsnr.JsonStringOf((structured, list(self), attributes), true_type=self)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

from typing import Iterator

import networkx as grph

import cell_tracking_BC.in_out.file.jsoner as jsnr
from cell_tracking_BC.type.cell import cell_t
from cell_tracking_BC.type.track.unstructured import TIME_POINT, unstructured_track_t


class unstructured_tracks_t(grph.DiGraph):
    @classmethod
    def NewFromJsonString(
        cls,
        jsoned: str,
        /,
    ) -> unstructured_tracks_t:
        """"""
        return cls(
            jsnr.ObjectFromJsonString(
                jsoned, builders={"cell_t": cell_t.NewFromJsonString}
            )
        )

    def AddTrackSegment(
        self,
        src_cell: cell_t,
        tgt_cell: cell_t,
        src_time_point: int,
        affinity: float,
        /,
    ) -> None:
        """"""
        time_point = {TIME_POINT: src_time_point}
        time_point_p_1 = {TIME_POINT: src_time_point + 1}
        self.add_node(src_cell, **time_point)
        self.add_node(tgt_cell, **time_point_p_1)
        self.add_edge(src_cell, tgt_cell, affinity=affinity)

    @property
    def track_iterator(self) -> Iterator[unstructured_track_t]:
        """"""
        for cells in grph.weakly_connected_components(self):
            track_view = self.subgraph(cells)
            # Copy or re-instantiation is necessary since the subgraph is a view
            yield unstructured_track_t(track_view)

    def AsJsonString(self) -> str:
        """"""
        return jsnr.JsonStringOf(self)

    def __eq__(self, other) -> bool:
        """"""
        if hasattr(other, "AsJsonString"):
            return self.AsJsonString() == other.AsJsonString()

        return False



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import (
    Callable,
    ClassVar,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import networkx as grph

import cell_tracking_BC.in_out.file.jsoner as jsnr
from cell_tracking_BC.in_out.text.logger import LOGGER
from cell_tracking_BC.standard.number import INFINITY_MINUS
from cell_tracking_BC.type.cell import cell_t, state_e
from cell_tracking_BC.type.track.single import single_track_t
from cell_tracking_BC.type.track.structured import (
    PerSingleTrackCells,
    cells_w_optional_time_points_h,
    per_single_track_cells_h,
    structured_track_t,
)
from cell_tracking_BC.type.track.unstructured import TIME_POINT, unstructured_track_t


# TODO: check if grph.descendants and grph.ancestors could be used in place of more complex code here
# TODO: cell iteration methods should have a geometric_mode: bool = False parameter to allow cell-state-independent
#     iterations


# Cannot be a dataclass due to in/out_degree declarations (which are only here to silence unfound attribute warnings)
class forking_track_t(structured_track_t, grph.DiGraph):
    """
    Affinities are stored as edge attributes
    """

    SINGLE_TRACK_LABEL: ClassVar[str] = "single_track_label"

    # Mutable Tuple[List[cell_t], List[cell_t]] for without and with time points lists
    _splitting_cells_cache: List[
        Optional[List[Union[cell_t, Tuple[cell_t, int]]]]
    ] = None

    in_degree: Callable[[cell_t], int]
    out_degree: Callable[[cell_t], int]

    def __init__(
        self, track: Union[grph.DiGraph, unstructured_track_t], **kwargs
    ) -> None:
        """"""
        if isinstance(track, unstructured_track_t):
            root, _ = track.RootCellWithTimePoint()
            leaves, _ = track.LeafCellsWithTimePoints()
            structured_track_t.__init__(self, root=root, leaves=leaves)
        # /!\ Otherwise, root and leaves (and the remaining attributes of structured_track_t) must be assigned later on.
        # Use case: NewFromJsonDescription
        grph.DiGraph.__init__(self, incoming_graph_data=track, **kwargs)

    @classmethod
    def NewFromUnstructuredTrack(
        cls, track: unstructured_track_t, next_single_track_label: int, /
    ) -> Tuple[forking_track_t, int]:
        """"""
        instance = cls(track)

        for label, leaf in enumerate(instance.leaves, start=next_single_track_label):
            # Adds attribute "forking_track_t.SINGLE_TRACK_LABEL" with value "label"
            # to leaf node indexed by "leaf" (not to the leaf itself).
            grph.set_node_attributes(
                instance,
                {leaf: label},
                name=forking_track_t.SINGLE_TRACK_LABEL,
            )

        return instance, next_single_track_label + instance.n_leaves

    @classmethod
    def NewFromJsonString(
        cls,
        jsoned: str,
        /,
    ) -> forking_track_t:
        """"""
        # See NewFromJsonDescription for why builders do not count one for structured_track_t
        return jsnr.ObjectFromJsonString(
            jsoned,
            builders={
                cls.__name__: cls.NewFromJsonDescription,
                "cell_t": cell_t.NewFromJsonString,
            },
        )

    @classmethod
    def NewFromJsonDescription(
        cls,
        description,
        /,
    ) -> forking_track_t:
        """"""
        structured, graph = description
        # See AsJsonString for why decoding happens here
        structured = structured_track_t.NewFromJsonString(structured)

        instance = cls(graph)

        # Do not use dtcl.asdict(self) since it recurses into dataclass instances which, if they extend a "container"
        # class like list or dict, will lose the contents.
        attributes = {
            _fld.name: getattr(structured, _fld.name)
            for _fld in dtcl.fields(structured)
        }
        for key, value in attributes.items():
            setattr(instance, key, value)

        return instance

    def AsJsonString(self) -> str:
        """"""
        structured = structured_track_t.AsJsonString(self)

        # grph.DiGraph(self): To hide this method, which would otherwise be infinitely recursively called
        return jsnr.JsonStringOf((structured, grph.DiGraph(self)), true_type=self)

    @property
    def cells(self) -> Union[Iterable, Iterator]:
        """"""
        return self.nodes

    @property
    def label(self) -> int:
        """"""
        raise RuntimeError("A Forking track has no unique single track label")

    @property
    def labels(self) -> Sequence[int]:
        """"""
        return tuple(self.TrackLabelWithLeaf(_lef, check=False) for _lef in self.leaves)

    @property
    def pruned_labels(self) -> Sequence[int]:
        """"""
        labels = tuple(
            (self.TrackLabelWithLeaf(_lef, check=False) for _lef in _lvs)
            for _lvs in (self.geometric_leaves, self.leaves)
        )

        return tuple(set(labels[0]).difference(labels[1]))

    @property
    def affinities(self) -> Tuple[float, ...]:
        """"""
        output = []

        for piece in self.Pieces():
            output.extend(piece.affinities)

        return tuple(output)

    def CellTimePoint(self, cell: cell_t) -> int:
        """"""
        return self.nodes[cell][TIME_POINT]

    def CellSuccessors(
        self, cell: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[Sequence[cell_t]]:
        """"""
        if (not check) or (cell in self):
            return tuple(
                _ngh
                for _ngh in self.neighbors(cell)
                if _ngh.state is not state_e.pruned
            )

        if tolerant_mode:
            return None

        raise ValueError(f"{cell}: Cell not in track")

    def CellDescendants(
        self,
        cell: cell_t,
        /,
        *,
        including_self: bool = True,
        geometric_mode: bool = False,
        check: bool = True,
        tolerant_mode: bool = False,
    ) -> Optional[Sequence[cell_t]]:
        """"""
        if (not check) or (cell in self):
            output = grph.descendants(self, cell)

            if geometric_mode:
                output = tuple(output)
            else:
                output = tuple(
                    _cll for _cll in output if _cll.state is not state_e.pruned
                )

            if including_self:
                output = (cell,) + output

            return output

        if tolerant_mode:
            return None

        raise ValueError(f"{cell}: Cell not in track")

    def ConfirmCellLineage(self, youngest: cell_t, oldest: cell_t, /) -> bool:
        """"""
        if youngest is oldest:
            if youngest in self:
                return True

            raise ValueError(f"{youngest}: Cell not in track")

        if (youngest in self) and (oldest in self):
            descendants = grph.descendants(self, youngest)

            return oldest in descendants

        raise ValueError(
            f"{youngest}/{oldest}: Cell track memberships={youngest in self}/{oldest in self}. "
            f"Expected=True/True"
        )

    def SplittingCells(
        self, /, *, with_time_point: bool = False
    ) -> Sequence[Union[cell_t, Tuple[cell_t, int]]]:
        """"""
        if with_time_point:
            which = 1
        else:
            which = 0
        if (self._splitting_cells_cache is None) or (
            self._splitting_cells_cache[which] is None
        ):
            if with_time_point:
                output = (
                    _rcd
                    for _rcd in self.nodes.data(TIME_POINT)
                    if self.out_degree(_rcd[0]) > 1
                )
            else:
                output = (_cll for _cll in self.nodes if self.out_degree(_cll) > 1)
            output = tuple(output)

            if self._splitting_cells_cache is None:
                self._splitting_cells_cache = [None, None]
            self._splitting_cells_cache[which] = output
        else:
            output = self._splitting_cells_cache[which]

        return output

    def MarkDividingCells(
        self,
        before_deaths: bool,
        /,
        *,
        division_responses: Dict[int, Optional[Sequence[float]]] = None,
        lower_bound: float = None,
        CombinedResponse: Callable[[Iterable], float] = max,
    ) -> None:
        """
        A division is set if the highest response at a splitting cell among the adjacent single tracks is above the
        threshold.
        """
        if before_deaths and not self._dead_marked:
            raise RuntimeError(
                'Division marking cannot be done in "before deaths" mode '
                "if dead cells have not been previously marked"
            )

        self._dividing_marked = True

        if (division_responses is None) or (lower_bound is None):
            for cell in self.SplittingCells():
                cell.state = state_e.dividing
            return

        for cell, time_point in self.SplittingCells(with_time_point=True):
            if cell.state in (state_e.dead, state_e.pruned):
                continue
            if before_deaths:
                ancestors = grph.ancestors(self, cell)
                if any(_cll.state is state_e.dead for _cll in ancestors):
                    continue

            sibling_labels = self.TrackLabelsContainingCell(cell)
            combined = CombinedResponse(
                division_responses[_lbl][time_point]
                if (division_responses[_lbl] is not None)
                and (division_responses[_lbl][time_point] is not None)
                else INFINITY_MINUS
                for _lbl in sibling_labels
            )
            if combined >= lower_bound:
                cell.state = state_e.dividing
            else:
                successors = self.CellSuccessors(cell)
                areas = tuple(_cll.area for _cll in successors)
                smallest = successors[areas.index(min(areas))]
                main_track = self.TrackLabelsContainingCell(cell)
                subtracks = self.TrackLabelsContainingCell(smallest)
                if subtracks.__len__() == 1:
                    subtracks = subtracks[0]
                LOGGER.info(
                    f"Full pruning of track {subtracks} "
                    f"(from time point {time_point + 1}) "
                    f"after splitting invalidation of track {main_track} "
                    f"at time point {time_point}"
                )
                for descendant in self.CellDescendants(smallest):
                    descendant.state = state_e.pruned
                    if descendant in self.leaves:
                        self.leaves = tuple(
                            _lef for _lef in self.leaves if _lef is not descendant
                        )

    @property
    def n_dividing_cells(self) -> int:
        """"""
        if not self._dividing_marked:
            raise RuntimeError("Dividing cells have not been marked yet")

        return sum(1 if _cll.state is state_e.dividing else 0 for _cll in self)

    def DividingCells(
        self,
        /,
        *,
        with_time_point: bool = False,
        per_single_track: bool = False,
    ) -> Union[cells_w_optional_time_points_h, per_single_track_cells_h]:
        """"""
        if not self._dividing_marked:
            raise RuntimeError("Dividing cells have not been marked yet")

        if with_time_point:
            output = (
                _rcd
                for _rcd in self.nodes.data(TIME_POINT)
                if _rcd[0].state is state_e.dividing
            )
        else:
            output = (_cll for _cll in self if _cll.state is state_e.dividing)
        # Note: PerSingleTrackCells does not accept an iterator
        output = tuple(output)

        if per_single_track:
            return PerSingleTrackCells(output, with_time_point, self)

        return output

    @property
    def segments_iterator(self) -> Iterator[Tuple[int, cell_t, cell_t, bool]]:
        """"""
        time_points = grph.get_node_attributes(self, TIME_POINT)

        for edge in self.edges:
            if all(_cll.state is not state_e.pruned for _cll in edge):
                time_point = time_points[edge[0]]
                is_last = edge[1] in self.leaves
                yield time_point, *edge, is_last

    def Pieces(
        self, /, *, from_cell: cell_t = None, with_time_point: int = None
    ) -> Sequence[single_track_t]:
        """"""
        output = []

        if from_cell is None:
            piece = [self.root]
            root_time_point = self.root_time_point
        else:
            if from_cell.state is state_e.pruned:
                raise ValueError(f"{from_cell}: Cell has been pruned")
            piece = [from_cell]
            root_time_point = with_time_point
        affinities = []

        while True:
            last_cell = piece[-1]

            if last_cell in self.leaves:
                neighbors = None
                n_neighbors = 0
            else:
                neighbors = self.CellSuccessors(last_cell, check=False)
                n_neighbors = neighbors.__len__()

            if n_neighbors == 0:
                label = self.TrackLabelWithLeaf(last_cell, check=False)
                output.append(
                    single_track_t.NewFromOrderedCells(
                        piece, affinities, root_time_point, label
                    )
                )
                break
            elif n_neighbors == 1:
                next_cell = neighbors[0]
                piece.append(next_cell)
                affinities.append(self[last_cell][next_cell]["affinity"])
            else:
                output.append(
                    single_track_t.NewFromOrderedCells(
                        piece, affinities, root_time_point, None
                    )
                )
                next_time_point = root_time_point + piece.__len__()
                for neighbor in neighbors:
                    pieces = self.Pieces(
                        from_cell=neighbor,
                        with_time_point=next_time_point,
                    )
                    for piece in pieces:
                        if piece[0] is neighbor:
                            # i.e. piece is a "root" piece, the others being located further down the subtrack tree
                            cells = (last_cell,) + tuple(piece)
                            affinity = self[last_cell][neighbor]["affinity"]
                            affinities = (affinity,) + piece.affinities
                            piece = single_track_t.NewFromOrderedCells(
                                cells,
                                affinities,
                                piece.root_time_point - 1,
                                piece.label,
                            )
                        output.append(piece)
                break

        return output

    def TrackLabelsContainingCell(
        self, cell: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[Sequence[int]]:
        """
        Returns empty tuple if cell has been pruned
        """
        if check and (cell not in self):
            if tolerant_mode:
                return None
            raise ValueError(f"{cell}: Cell not in track")

        if cell.state is state_e.pruned:
            return ()

        descendants = self.CellDescendants(cell)
        leaves = set(self.leaves).intersection(descendants)
        output = tuple(self.TrackLabelWithLeaf(_lef, check=False) for _lef in leaves)

        assert output.__len__() > 0

        return output

    def TrackLabelWithLeaf(
        self, leaf: cell_t, /, *, check: bool = True, tolerant_mode: bool = False, geometric_mode: bool = False,
    ) -> Optional[int]:
        """
        TrackLabelWithLeaf: Implicitly, it is SingleTrackLabelWithLeaf
        """
        if geometric_mode:
            leaves = self.geometric_leaves
        else:
            leaves = self.leaves
        if (not check) or (leaf in leaves):
            label = self.nodes[leaf].get(forking_track_t.SINGLE_TRACK_LABEL)
            if label is None:
                # The track has been pruned from after this "logical" leaf. All the labels of the descendant, geometric
                # leaves are therefore unused. Their minimum will be picked as the label.
                descendants = grph.descendants(self, leaf)
                geometric_leaves = set(self.geometric_leaves).intersection(descendants)
                label = min(
                    self.nodes[_lef][forking_track_t.SINGLE_TRACK_LABEL]
                    for _lef in geometric_leaves
                )

            return label

        if tolerant_mode:
            return None

        raise ValueError(f"{leaf}: Not a leaf cell")

    def TrackWithLeaf(
        self, leaf: cell_t, /, *, check: bool = True, tolerant_mode: bool = False,
        geometric_mode: bool = False,
    ) -> Optional[single_track_t]:
        """
        TrackWithLeaf: Implicitly, it is SingleTrackWithLeaf
        """
        if geometric_mode:
            leaves = self.geometric_leaves
        else:
            leaves = self.leaves
        # TODO: currently, geometric mode only accounts for leaves, not root
        if (not check) or (leaf in leaves):
            cells = grph.shortest_path(self, source=self.root, target=leaf)
            affinities = tuple(
                self[cells[_idx]][cells[_idx + 1]]["affinity"]
                for _idx in range(cells.__len__() - 1)
            )
            label = self.TrackLabelWithLeaf(leaf, check=False, geometric_mode=geometric_mode)
            output = single_track_t.NewFromOrderedCells(
                cells, affinities, self.root_time_point, label
            )

            return output

        if tolerant_mode:
            return None

        raise ValueError(f"{leaf}: Not a leaf cell")

    def AsSingleTrack(self) -> single_track_t:
        """"""
        output = [self.root]

        affinities = []
        while True:
            last_cell = output[-1]

            if last_cell in self.leaves:
                neighbors = None
                n_neighbors = 0
            else:
                neighbors = self.CellSuccessors(last_cell, check=False)
                n_neighbors = neighbors.__len__()

            if n_neighbors == 0:
                label = self.TrackLabelWithLeaf(last_cell, check=False)
                break
            elif n_neighbors == 1:
                next_cell = neighbors[0]
                output.append(next_cell)
                affinity = self[last_cell][next_cell]["affinity"]
                affinities.append(affinity)
            else:
                raise ValueError(
                    f"Attempt to convert the forking track with root {self.root} and "
                    f"{self.n_leaves} leaves into a single track"
                )

        output = single_track_t.NewFromOrderedCells(
            output, affinities, self.root_time_point, label
        )

        return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from operator import itemgetter as ItemAt
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Union,
)

import cell_tracking_BC.in_out.file.jsoner as jsnr
from cell_tracking_BC.in_out.text.logger import LOGGER
from cell_tracking_BC.standard.number import INFINITY_MINUS
from cell_tracking_BC.standard.uid import Identity
from cell_tracking_BC.type.cell import cell_t, state_e


feature_filtering_h = Callable[..., Optional[Sequence[Any]]]
per_single_track_feature_h = Dict[int, Optional[Sequence[Any]]]
cells_wo_time_points = Sequence[cell_t]
cells_w_time_points_h = Sequence[Tuple[cell_t, int]]
cells_w_optional_time_points_h = Union[cells_wo_time_points, cells_w_time_points_h]
per_single_track_cells_h = Dict[int, cells_w_optional_time_points_h]
per_single_track_cell_h = Dict[int, Optional[Union[cell_t, Tuple[cell_t, int]]]]


# TODO: check if grph.descendants and grph.ancestors could be used in place of more complex code here
# TODO: cell iteration methods should have a geometric_mode: bool = False parameter to allow cell-state-independent
#     iterations


class single_additional_p(Protocol):
    label: int
    length: int

    def CellFeature(self, feature: str, /, *, geometric_mode: bool = False) -> Sequence[Any]:
        """"""

    def __getitem__(self, item):
        """"""


@dtcl.dataclass(repr=False, eq=False)
class structured_track_t:
    """
    Adding Iterable to the class inheritance silences warnings at "cell in self". Unfortunately, forking tracks become
    un-instantiable for they "lack" an __iter__ method.

    root, leaves: "logical" versions, i.e. accounting for pruning. Geometric versions are geometric_root and
    geometric_leaves.
    """

    root: cell_t = None
    leaves: Sequence[cell_t] = None
    # TODO: take advantage of these new geometric attr
    geometric_root: cell_t = dtcl.field(init=False, default=None)
    geometric_leaves: Tuple[cell_t] = dtcl.field(init=False, default=None)
    # str: feature name; int: single track label
    features: Dict[str, per_single_track_feature_h] = dtcl.field(
        init=False, default_factory=dict
    )
    _dividing_marked: bool = dtcl.field(init=False, default=False)
    _dead_marked: bool = dtcl.field(init=False, default=False)

    def __post_init__(self) -> None:
        """"""
        self.geometric_root = self.root
        self.geometric_leaves = tuple(self.leaves)

    @classmethod
    def NewFromJsonString(
        cls,
        jsoned: str,
        /,
    ) -> structured_track_t:
        """"""
        return jsnr.ObjectFromJsonString(
            jsoned,
            builders={
                cls.__name__: cls.NewFromJsonDescription,
                "cell_t": cell_t.NewFromJsonString,
            },
        )

    @classmethod
    def NewFromJsonDescription(
        cls,
        description,
        /,
    ) -> structured_track_t:
        """"""
        instance = cls.__new__(cls)
        for key, value in description.items():
            setattr(instance, key, value)

        return instance

    def AsJsonString(self) -> str:
        """"""
        # Forcing true_type guarantees that derived classes will not "cover" its value with their own type. Otherwise,
        # the decoding of a true structured_track_t instance will be incorrectly routed to the derived classes.
        return jsnr.JsonStringOf(self, true_type=structured_track_t)

    @property
    def cells(self) -> Union[Iterable, Iterator]:
        """"""
        raise NotImplementedError

    @property
    def labels(self) -> Sequence[int]:
        """
        Single track labels
        """
        raise NotImplementedError

    @property
    def labels_as_str(self) -> str:
        """
        Single track labels
        """
        return "+".join(str(_lbl) for _lbl in sorted(self.labels))

    @property
    def pruned_labels(self) -> Sequence[int]:
        """"""
        raise NotImplementedError

    @property
    def n_leaves(self) -> int:
        """"""
        return self.leaves.__len__()

    @property
    def lengths(self) -> Sequence[int]:
        """
        Segment-wise, not node-wise
        """
        return tuple(_ltp - self.root_time_point for _ltp in self.leaves_time_points)

    @property
    def root_time_point(self) -> int:
        """"""
        return self.CellTimePoint(self.root)

    @property
    def leaves_time_points(self) -> Sequence[int]:
        """"""
        return tuple(self.CellTimePoint(_lef) for _lef in self.leaves)

    def CellTimePoint(self, cell: cell_t) -> int:
        """"""
        raise NotImplementedError

    def CellSuccessors(
        self, cell: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[Sequence[cell_t]]:
        """
        Accounts for pruning
        """
        raise NotImplementedError

    def CellDescendants(
        self,
        cell: cell_t,
        /,
        *,
        including_self: bool = True,
        geometric_mode: bool = False,
        check: bool = True,
        tolerant_mode: bool = False,
    ) -> Optional[Sequence[cell_t]]:
        """"""
        raise NotImplementedError

    def ConfirmCellLineage(self, youngest: cell_t, oldest: cell_t, /) -> bool:
        """
        Including youngest is oldest
        """
        raise NotImplementedError

    def SplittingCells(
        self, /, *, with_time_point: bool = False
    ) -> Sequence[Union[cell_t, Tuple[cell_t, int]]]:
        """"""
        raise NotImplementedError

    def MarkDividingCells(
        self,
        before_deaths: bool,
        /,
        *,
        division_responses: Dict[int, Optional[Sequence[float]]] = None,
        lower_bound: float = None,
        CombinedResponse: Callable[[Iterable], float] = max,
    ) -> None:
        """"""
        raise NotImplementedError

    @property
    def n_dividing_cells(self) -> int:
        """"""
        raise NotImplementedError

    def DividingCells(
        self,
        /,
        *,
        with_time_point: bool = False,
        per_single_track: bool = False,
    ) -> Union[cells_w_optional_time_points_h, per_single_track_cells_h]:
        """"""
        raise NotImplementedError

    def DivisionTimePoints(self) -> Dict[int, Optional[Sequence[int]]]:
        """
        Note: There is no sense in returning not-per-single-track time points
        """
        if not self._dividing_marked:
            raise RuntimeError("Dividing cells have not been marked yet")

        dividing = self.DividingCells(with_time_point=True, per_single_track=True)

        if dividing.__len__() > 0:
            output = {
                _key: None if _val is None else tuple(map(ItemAt(1), _val))
                for _key, _val in dividing.items()
            }

            return output

        return {}  # Used to be (-1,), then None

    def MarkDeadCells(
        self,
        death_responses: Dict[int, Optional[Sequence[float]]],
        lower_bound: float,
        after_divisions: bool,
        /,
        *,
        CombinedResponse: Callable[[Iterable], float] = max,
        called_from_educated_code: bool = False,
    ) -> None:
        """
        A death is set if the highest response at a cell among the adjacent single tracks is above the threshold.

        called_from_educated_code: A track might be entirely pruned out here. Since it cannot remove itself from its
        referring objects, this method must be called from a piece of code that can do it if needed. This parameter is
        meant to prevent an "accidental" call from some "uneducated" piece of code.
        """
        if after_divisions and not self._dividing_marked:
            raise RuntimeError(
                'Death marking cannot be done in "after divisions" mode '
                "if dividing cells have not been previously marked"
            )
        if not called_from_educated_code:
            raise ValueError(
                f"{self.MarkDeadCells.__name__}: Must be called from a piece of code handling full pruning"
            )

        self._dead_marked = True

        if (death_responses is None) or (lower_bound is None):
            return

        if after_divisions:
            dividing_cells = self.DividingCells()
        else:
            dividing_cells = ()

        # This loop must not prune cells or modify leaves since it could change the sibling labels to labels of leaves
        # that have been pruned in a prior step (e.g. when validating divisions). For example, behind a cell are 2
        # leaves with labels 1 and 2. Suppose that the leaf with label 1 has been pruned in a prior step. So the cell
        # has 2 as its sibling labels. So has the cell before it. If the cell is marked dead, and the leaf with label 2
        # is pruned, the previous cell will now receive the label min(1,2), which is different from what it had.
        dead_cells = []
        for cell in self.cells:
            if cell.state in (state_e.dividing, state_e.pruned):
                continue
            if after_divisions and any(
                self.ConfirmCellLineage(cell, _dvd) for _dvd in dividing_cells
            ):
                continue

            time_point = self.CellTimePoint(cell)
            sibling_labels = self.TrackLabelsContainingCell(cell)
            combined = CombinedResponse(
                death_responses[_lbl][time_point]
                if (death_responses[_lbl] is not None)
                and (death_responses[_lbl][time_point] is not None)
                else INFINITY_MINUS
                for _lbl in sibling_labels
            )
            if combined >= lower_bound:
                # Note: The cell can be a leaf (see new_leaves below)
                dead_cells.append(cell)

        # Do not log inside the loop on dead_cells since it does not update leaves (update is done after the loop),
        # leaving the track in a transient state not suitable for TrackLabelsContainingCell.
        summary = {}
        max_time_point = self.lengths[-1]
        for cell in dead_cells:
            tracks = self.TrackLabelsContainingCell(cell)
            if tracks.__len__() == 1:
                tracks = tracks[0]
            time_point = self.CellTimePoint(cell)
            summary[tracks] = min(summary.get(tracks, max_time_point), time_point)
        for tracks, time_point in summary.items():
            LOGGER.info(
                f"Partial pruning of track {tracks} after dead cell at time point {time_point}"
            )

        invalid_leaves = []
        for cell in dead_cells:
            if cell.state is state_e.pruned:
                # It has been pruned below as a descendant
                continue

            cell.state = state_e.dead

            for descendant in self.CellDescendants(cell, including_self=False):
                descendant.state = state_e.pruned
                if descendant in self.leaves:
                    invalid_leaves.append(descendant)

        # 1. Do not use dead_cells here since it may contain dead cells after other dead cells
        # 2. Cells marked dead can already be leaves; They must not be included in new_leaves then
        new_leaves = tuple(
            _cll
            for _cll in self.cells
            if (_cll.state is state_e.dead) and (_cll not in self.leaves)
        )
        self.leaves = tuple(set(self.leaves).difference(invalid_leaves)) + new_leaves

    def DeadCells(
        self,
        sequence_length: int,
        /,
        *,
        with_time_point: bool = False,
        with_living_leaves: bool = False,
        per_single_track: bool = False,
    ) -> Union[cells_w_optional_time_points_h, per_single_track_cell_h]:
        """"""
        if not self._dead_marked:
            raise RuntimeError("Dead cells have not been marked yet")

        output = []

        for leaf in self.leaves:
            time_point = self.CellTimePoint(leaf)
            if (is_dead := leaf.state is state_e.dead) or (
                with_living_leaves and (time_point < sequence_length - 1)
            ):
                if with_time_point:
                    if is_dead:
                        output.append((leaf, time_point))
                    else:
                        output.append((leaf, -time_point))
                else:
                    output.append(leaf)

        if per_single_track:
            return PerSingleTrackCells(output, with_time_point, self, unique_cell=True)

        return output

    def DeathTimePoints(
        self,
        sequence_length: int,
        /,
        *,
        with_living_leaves: bool = False,
    ) -> Optional[Dict[int, Optional[int]]]:
        """
        Note: There is no sense in returning not-per-single-track time points
        """
        if not self._dead_marked:
            raise RuntimeError("Dead cells have not been marked yet")

        dead = self.DeadCells(
            sequence_length,
            with_time_point=True,
            with_living_leaves=with_living_leaves,
            per_single_track=True,
        )

        if dead.__len__() > 0:
            output = {
                _key: None if _val is None else _val[1] for _key, _val in dead.items()
            }

            return output

        return {}  # Used to be (-1,), then None

    @property
    def segments_iterator(self) -> Iterator[Tuple[int, cell_t, cell_t, bool]]:
        """"""
        raise NotImplementedError

    def Pieces(
        self, /, *, from_cell: cell_t = None, with_time_point: int = None
    ) -> Sequence[structured_track_t]:
        """"""
        raise NotImplementedError

    @property
    def single_tracks_iterator(
        self,
    ) -> Iterator[Union[structured_track_t, single_additional_p]]:
        """"""
        for leaf in self.leaves:
            yield self.TrackWithLeaf(leaf, check=False)

    @property
    def geometric_single_tracks_iterator(
        self,
    ) -> Iterator[Union[structured_track_t, single_additional_p]]:
        """"""
        for leaf in self.geometric_leaves:
            yield self.TrackWithLeaf(leaf, check=False, geometric_mode=True)

    def TrackLabelsContainingCell(
        self, cell: cell_t, /, *, check: bool = True, tolerant_mode: bool = False
    ) -> Optional[Sequence[int]]:
        """"""
        raise NotImplementedError

    def TrackLabelWithLeaf(
        self, leaf: cell_t, /, *, check: bool = True, tolerant_mode: bool = False, geometric_mode: bool = False,
    ) -> Optional[int]:
        """"""
        raise NotImplementedError

    def TrackWithLeaf(
        self,
        leaf: cell_t,
        /,
        *,
        check: bool = True,
        tolerant_mode: bool = False,
        geometric_mode: bool = False,
    ) -> Optional[structured_track_t]:
        """
        Actually, a single track
        """
        raise NotImplementedError

    def AsSingleTrack(self) -> structured_track_t:
        """"""
        raise NotImplementedError

    def IsFullyPruned(self) -> bool:
        """"""
        return self.root.state is state_e.pruned

    def AddFeature(
        self,
        name: str,
        cell_feature_names: Sequence[str],
        FeatureComputation: feature_filtering_h,
        /,
        *args,
        geometric_mode: bool = False,
        **kwargs,
    ) -> None:
        """
        /!\ There is no already-existing check
        """
        self.features[name] = {}

        per_single_track = self.features[name]
        if geometric_mode:
            iterator = self.geometric_single_tracks_iterator
        else:
            iterator = self.single_tracks_iterator
        for single in iterator:
            all_series = [single.CellFeature(_ftr, geometric_mode = geometric_mode) for _ftr in cell_feature_names]
            feature = FeatureComputation(
                *all_series, *args, track_label=single.label, **kwargs
            )
            per_single_track[single.label] = feature

    def __repr__(self) -> str:
        """"""
        return Identity(self)

    def __str__(self) -> str:
        """"""
        if hasattr(self, "nodes"):
            cells = self.nodes
        else:
            cells = self
        cell_labels = tuple(_cll.label for _cll in cells)

        return (
            f"{repr(self)}:\n"
            f"    {self.labels=}\n"
            f"    {self.root_time_point=}\n"
            f"    {self.leaves_time_points=}\n"
            f"    {self.lengths=}\n"
            f"    {cell_labels}"
        )

    def __eq__(self, other) -> bool:
        """"""
        if hasattr(other, "AsJsonString"):
            return self.AsJsonString() == other.AsJsonString()

        return False

    def __hash__(self) -> int:
        """"""
        return hash((self.root, *self.leaves))


def PerSingleTrackCells(
    linear: cells_w_optional_time_points_h,
    has_time_point: bool,
    track: structured_track_t,
    /,
    *,
    unique_cell: bool = False,
) -> Union[per_single_track_cell_h, per_single_track_cells_h]:
    """"""
    output = {}

    for single in track.single_tracks_iterator:
        if has_time_point:
            per_single = filter(lambda _rcd: _rcd[0] in single, linear)
            per_single = sorted(per_single, key=ItemAt(1))
        else:
            per_single = tuple(filter(lambda _cll: _cll in single, linear))

        if (n_cells := per_single.__len__()) > 0:
            if unique_cell:
                if n_cells > 1:
                    raise RuntimeError(
                        f"{n_cells}/{per_single}: Invalid number of cells in unique-cell mode "
                        f"for single track {single.label}"
                    )

                output[single.label] = per_single[0]
            else:
                output[single.label] = per_single
        elif unique_cell:
            output[single.label] = None
        else:
            output[single.label] = ()

    return output



# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Any, Dict, Sequence, Tuple, TypeVar, Union


annotation_h = TypeVar("annotation_h")
# "some_text" or (("text1", properties1), ("text2", properties2)...)
cell_annotation_h = Union[str, Sequence[Tuple[str, Dict[str, Any]]]]



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from pathlib import Path as path_t
from typing import Any

import imageio as mgio
import numpy as nmpy


array_t = nmpy.ndarray


def FrameByIMAGEIO(path: path_t, /) -> array_t:
    """"""
    return mgio.imread(path)


try:
    # noinspection PyPackageRequirements
    import itk as ittk

    def FrameByITK(path: path_t, /) -> array_t:
        """
        Also works for sequences
        """
        itk_image = ittk.imread(str(path))
        output = ittk.array_from_image(itk_image)

        return output

    def SaveFrameWithITK(frame: array_t, path: path_t, /) -> None:
        """"""
        itk_np_view = ittk.image_view_from_array(frame)
        ittk.imwrite(itk_np_view, str(path))

except ModuleNotFoundError:

    def RaiseMissingITKException(*_, **__) -> Any:
        """"""
        raise RuntimeError("ITK (https://itk.org/) is not available")

    FrameByITK = RaiseMissingITKException
    SaveFrameWithITK = RaiseMissingITKException



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Callable, Optional, Sequence, Tuple, Union

import matplotlib.lines as lins
import matplotlib.pyplot as pypl
import numpy as nmpy
from matplotlib.collections import PatchCollection as patches_t
from matplotlib.legend_handler import HandlerTuple as tuple_handler_t
from matplotlib.patches import Rectangle as mpl_rectangle_t

from cell_tracking_BC.in_out.file.archiver import archiver_t
from cell_tracking_BC.in_out.graphics.dbe.matplotlib.d_2 import figure_t
from cell_tracking_BC.in_out.graphics.event import rectangle_t


rectangle_info_h = Tuple[int, float, float, bool, bool]
ValueTransform_h = Optional[Union[float, Callable[[float], float]]]


DIVISION_MARKER = "3"
DEATH_MARKER = "x"
DIVISION_MARKER_SIZE = 60
DEATH_MARKER_SIZE = 30
N_SCORE_LEGEND_SHADES = 8

_RECTANGLE_TYPE_DIV = 0
_RECTANGLE_TYPE_DEA = 1
_RECTANGLE_TYPE_MIXED = 2

_rectangle = mpl_rectangle_t((0.0, 0.0), 1.0, 1.0)
_rectangle.set_facecolor("None")
_NO_COLOR_MATPLOTLIB = _rectangle.get_facecolor()


def PlotDivisionAndDeathEventResponses(
    events: Sequence[Tuple[str, int, Sequence[float]]],
    rectangles: Sequence[rectangle_t],
    track_height: int,
    sequence_length: int,
    max_label: int,
    zero_is_black: bool = True,
    show_as_barplot: bool = False,
    figure_name: str = "division-and-death-responses",
    prepare_only: bool = False,
    interactively: bool = True,
    in_main_thread: bool = True,
    archiver: archiver_t = None,
) -> Optional[figure_t]:
    """"""
    output = None

    mpl_rectangles = []
    for rectangle in rectangles:
        *geometry, edge_color, face_color = rectangle.Details()
        if face_color is None:
            face_color = _NO_COLOR_MATPLOTLIB

        mpl_rectangle = mpl_rectangle_t(*geometry)
        mpl_rectangle.set_edgecolor(edge_color)
        mpl_rectangle.set_facecolor(face_color)

        mpl_rectangles.append(mpl_rectangle)

    edge_colors = tuple(_rct.get_edgecolor() for _rct in mpl_rectangles)
    face_colors = tuple(_rct.get_facecolor() for _rct in mpl_rectangles)
    patches = patches_t(mpl_rectangles, edgecolors=edge_colors, facecolors=face_colors)

    figure, axes = figure_t.NewFigureAndAxes()

    axes.add_collection(patches)
    if show_as_barplot:
        color = "k"
    elif zero_is_black:
        color = "w"
    else:
        color = "k"
    for what, size, where in events:
        axes.scatter(*where, c=color, marker=what, s=size)
    _SetAxesProperties(axes, sequence_length, max_label, track_height)
    _AddLegend(axes, zero_is_black)

    figure.Archive(name=figure_name, archiver=archiver)
    if prepare_only:
        output = figure
    else:
        figure.Show(interactively=interactively, in_main_thread=in_main_thread)

    return output


def _SetAxesProperties(
    axes: pypl.Axes, sequence_length: int, max_label: int, track_height: int, /
) -> None:
    """"""
    axes.set_xlim(left=0, right=sequence_length)
    axes.set_ylim(bottom=0, top=track_height * max_label + 1)

    if track_height == 2:
        offset = -0.5
    else:
        offset = 0.0
    positions = range(1, max_label + 1, max(1, int(round(max_label / 10))))
    axes.yaxis.set_ticks(track_height * nmpy.array(positions) + offset)
    axes.yaxis.set_ticklabels(positions)


def _AddLegend(axes: pypl.Axes, zero_is_black: bool, /) -> None:
    """"""
    shades = nmpy.linspace(0.0, 1.0, num=N_SCORE_LEGEND_SHADES)

    if zero_is_black:
        DivisionColor = lambda _shd: (0.0, 0.0, _shd)
        DeathColor = lambda _shd: (_shd, 0.0, 0.0)
    else:
        DivisionColor = lambda _shd: (1.0 - _shd, 1.0 - _shd, 1.0)
        DeathColor = lambda _shd: (1.0, 1.0 - _shd, 1.0 - _shd)
    score_legends = [
        [
            mpl_rectangle_t(
                (0, 0),
                3.0,
                5.0,
                edgecolor=_NO_COLOR_MATPLOTLIB,
                facecolor=_Clr(_shd),
            )
            for _shd in shades
        ]
        for _Clr in (DivisionColor, DeathColor)
    ]

    event_legends = [
        lins.Line2D(
            (),
            (),
            color="k",
            marker=_mrk,
            markersize=_sze,
            linestyle="None",
        )
        for _mrk, _sze in (
            (DIVISION_MARKER, DIVISION_MARKER_SIZE // 5),
            (DEATH_MARKER, DEATH_MARKER_SIZE // 5),
        )
    ]

    axes.legend(
        handles=(*score_legends, *event_legends),
        labels=("Division Score", "Death Score", "Division", "Death (pattern)"),
        loc="right",
        bbox_to_anchor=(1.3, 0.5),
        handler_map={list: tuple_handler_t(ndivide=None, pad=0)},
    )



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import cast

from cell_tracking_BC.in_out.graphics.type.annotation import annotation_h
from cell_tracking_BC.in_out.graphics.type.axes import axes_2d_t as axes_t
from cell_tracking_BC.in_out.graphics.type.context import context_t
from cell_tracking_BC.in_out.graphics.type.event import event_h
from cell_tracking_BC.in_out.graphics.type.figure import figure_t
from cell_tracking_BC.in_out.graphics.type.s_viewer_2d import s_viewer_2d_t
from cell_tracking_BC.in_out.graphics.type.t_viewer_2d import t_viewer_2d_t
from cell_tracking_BC.type.sequence import sequence_t


@dtcl.dataclass(repr=False, eq=False)
class t_explorer_2d_t:

    figure: figure_t
    t_viewer: t_viewer_2d_t
    s_viewer: s_viewer_2d_t

    annotation: annotation_h = dtcl.field(init=False, default=None)

    dbe: context_t = None

    @classmethod
    def NewForSequence(
        cls,
        sequence: sequence_t,
        dbe: context_t,
        /,
        *,
        version: str = None,
        with_segmentation: bool = True,
        with_cell_labels: bool = True,
        with_track_labels: bool = True,
        with_ticks: bool = True,
        with_colorbar: bool = True,
        mode: str = "forking",
    ) -> t_explorer_2d_t:
        """
        mode: see t_viewer_2d_t.NewForTracks
        Annotation-on-hover code adapted from:
        https://stackoverflow.com/questions/7908636/how-to-add-hovering-annotations-in-matplotlib
        Answered by ImportanceOfBeingErnest on Nov 7 '17 at 20:23
        Edited by waterproof on Aug 12 '19 at 20:08
        """
        figure, two_axes = dbe.figure_2d_t.NewFigureAndAxes(n_cols=2)
        axes_track = cast(axes_t, two_axes[0])
        axes_viewer = cast(axes_t, two_axes[1])

        t_viewer = t_viewer_2d_t.NewForTracks(
            sequence.tracks, dbe, mode=mode, in_axes=axes_track
        )
        s_viewer = dbe.s_viewer_2d_t.NewForAllStreams(
            sequence,
            dbe,
            version=version,
            with_segmentation=with_segmentation,
            with_cell_labels=with_cell_labels,
            with_track_labels=with_track_labels,
            in_axes=axes_viewer,
            with_ticks=with_ticks,
            with_colorbar=with_colorbar,
        )
        # Cannot set conversion function here since it creates an import cycle
        # s_viewer.SetConversionToAnnotatedVolume(AsAnnotatedVolume)

        instance = cls(
            figure=figure,
            t_viewer=t_viewer,
            s_viewer=s_viewer,
            dbe=dbe,
        )

        annotation = t_viewer.axes.PlotAnnotation(
            "",
            xy=(0, 0),
            xytext=(20, 20),
            textcoords="offset points",
            bbox={"boxstyle": "round", "fc": "c"},
            arrowprops={"arrowstyle": "-"},
        )
        dbe.SetVisibility(annotation, False)

        instance.annotation = annotation

        figure.ActivateEvent("button_press_event", instance._OnButtonPress)

        figure.ActivateTightLayout(h_pad=0.05)

        return instance

    def _UpdateLinkedAnnotation(
        self,
        event: event_h,
        /,
    ) -> None:
        """"""
        raise NotImplementedError

    def _OnButtonPress(
        self,
        event: event_h,
        /,
    ) -> None:
        """"""
        if self.dbe.IsTargetOfEvent(self.t_viewer.axes, event):
            self._UpdateLinkedAnnotation(event)
        elif self.dbe.IsTargetOfEvent(self.s_viewer.axes, event):
            pass
        elif self.dbe.IsVisible(self.annotation):
            self.dbe.SetVisibility(self.annotation, False)
            self.figure.Update()



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from typing import Optional, Sequence, Union

from cell_tracking_BC.standard.issue import ISSUE_SEPARATOR
from cell_tracking_BC.standard.number import MAX_INT
from cell_tracking_BC.type.cell import state_e
from cell_tracking_BC.type.track.forking import forking_track_t
from cell_tracking_BC.type.track.single import single_track_t
from cell_tracking_BC.type.track.unstructured import unstructured_track_t


structured_track_h = Union[single_track_t, forking_track_t]
any_track_h = Union[unstructured_track_t, structured_track_h]


# def DivisionTimePoints(
#     dividing_cells: Sequence[Tuple[cell_t, int]], /
# ) -> Tuple[Optional[Sequence[int]], int]:
#     """"""
#     division_time_points = tuple(_elm[1] for _elm in dividing_cells)
#     if division_time_points.__len__() > 0:
#         last_div_frm = division_time_points[-1] + 1
#     else:
#         division_time_points = None  # Used to be (-1,)
#         last_div_frm = 0
#
#     return division_time_points, last_div_frm


def BasicTrackIssues(
    track: structured_track_h,
    /,
    *,
    root_time_point_interval: Optional[Sequence[Optional[int]]] = None,
    leaves_time_point_interval: Optional[Sequence[Optional[int]]] = None,
    min_length: int = 0,
    max_n_children: int = 2,
    can_touch_border: bool = False,
) -> Optional[Sequence[str]]:
    """
    All parameters: any limit can be ignored by setting it to None
    All intervals are inclusive.
    leaf_time_point_intervals and min_lengths: first element is for the shortest branch, the second is for the longest.
        For single tracks, both are the same.
    min_length: min edge-wise length, inclusive, of shortest single track
    max_n_children: inclusive
    """
    output = []

    if track.root.state is state_e.dead:
        output.append('Root cell has a "dead" state')

    mini, maxi = _IntervalWithDefaults(root_time_point_interval, 0, MAX_INT)
    if not (mini <= track.root_time_point <= maxi):
        output.append(
            f"{track.root_time_point}{ISSUE_SEPARATOR}Invalid root time point. Expected={mini}..{maxi}."
        )

    min_ltp = min(track.leaves_time_points)
    max_ltp = max(track.leaves_time_points)
    mini, maxi = _IntervalWithDefaults(leaves_time_point_interval, 0, MAX_INT)
    for time_point, which in zip(
        (min_ltp, max_ltp), ("shortest", "longest")
    ):
        if not (mini <= time_point <= maxi):
            output.append(
                f"{time_point}{ISSUE_SEPARATOR}Invalid leaf time point of {which} branch. Expected={mini}..{maxi}."
            )

    if track.lengths[0] < min_length:
        output.append(
            f"{track.lengths[0]}{ISSUE_SEPARATOR}Invalid (edge-wise) length of shortest branch. Expected>={min_length}."
        )

    if isinstance(track, forking_track_t):
        for cell in track.nodes:
            if (n_children := track.out_degree(cell)) > max_n_children:
                output.append(
                    f"C{cell.label}T{track.CellTimePoint(cell)}{ISSUE_SEPARATOR}"
                    f"{n_children} successors. Expected=0..{max_n_children}."
                )

    if not can_touch_border:
        for cell in track.cells:
            if cell.touches_border:
                output.append(
                    f"C{cell.label}T{track.CellTimePoint(cell)}{ISSUE_SEPARATOR}Touches frame border"
                )

    if output.__len__() == 0:
        output = None

    return output


def _IntervalWithDefaults(
    interval: Optional[Sequence[Optional[int]]], default_min: int, default_max: int, /
) -> Sequence[int]:
    """"""
    if interval is None:
        return default_min, default_max

    tvl_min, tvl_max = interval
    if tvl_min is None:
        tvl_min = default_min
    if tvl_max is None:
        tvl_max = default_max

    return tvl_min, tvl_max



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import string as strg
from typing import Any, Tuple


_ASCII_OF = (
    strg.ascii_lowercase + strg.digits + "".join(chr(_asc) for _asc in range(192, 256))
)


def Identity(obj: Any, /) -> str:
    """"""
    as_str = str(hash(obj))
    uid = (
        _ASCII_OF[int(as_str[c_idx : (c_idx + 2)])]
        for c_idx in range(0, as_str.__len__(), 2)
    )
    uid = "".join(uid)

    return f"{obj.__class__.__name__.upper()}.{uid}"


def AlphaColumnFromIndex(index: int) -> str:
    """
    index: 0-based
    """
    return AlphaColumnFromLabel(index + 1)


def AlphaColumnFromLabel(label: int) -> str:
    """
    label: 1-based

    Adapted from: https://stackoverflow.com/questions/48983939/convert-a-number-to-excel-s-base-26
    Answer: poke, Feb 26 '18 at 9:00 (https://stackoverflow.com/a/48984697)
    """

    def _DivmodExcel(number: int) -> Tuple[int, int]:
        quotient, remainder = divmod(number, 26)
        if remainder > 0:
            return quotient, remainder
        else:
            return quotient - 1, remainder + 26

    output = []

    while label > 0:
        label, digit = _DivmodExcel(label)
        output.append(strg.ascii_uppercase[digit - 1])

    return "".join(reversed(output))



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from enum import Enum as enum_t
from typing import Any, Optional, Sequence, Tuple, Union

import numpy as nmpy
import scipy.ndimage as ndimage_t

import cell_tracking_BC.in_out.file.jsoner as jsnr
from cell_tracking_BC.type.compartment import compartment_t
from cell_tracking_BC.type.cytoplasm import cytoplasm_t
from cell_tracking_BC.type.nucleus import nucleus_t


array_t = nmpy.ndarray
nucleus_or_nuclei_h = Union[nucleus_t, Tuple[nucleus_t, nucleus_t]]


class state_e(enum_t):
    unknown = None
    pruned = -1
    living = 1
    dividing = 2
    dead = 0

    def IsActive(self, /, *, strict_mode: bool = False) -> bool:
        """"""
        if strict_mode:
            return self not in (state_e.dead, state_e.pruned, state_e.unknown)

        return self not in (state_e.dead, state_e.pruned)

    def IsAlive(self, /, *, strict_mode: bool = False) -> bool:
        """"""
        if strict_mode:
            return self in (state_e.living, state_e.dividing)

        return self in (state_e.living, state_e.dividing, state_e.unknown)


@dtcl.dataclass(repr=False, eq=False)
class cell_t(compartment_t):

    label: Any = None
    nucleus: Optional[nucleus_or_nuclei_h] = None
    cytoplasm: Optional[cytoplasm_t] = None
    # Inherited from compartment_t (among others): features
    state: state_e = state_e.unknown

    @classmethod
    def NewFromMap(cls, _: array_t, /, *, is_plain: bool = True) -> compartment_t:
        """"""
        raise RuntimeError(
            f"{cell_t.NewFromMap.__name__}: Not meant to be called from class {cell_t.__name__}; "
            f"Use {cell_t.NewFromMaps.__name__} instead"
        )

    @classmethod
    def NewFromMaps(
        cls,
        label: Any,
        /,
        *,
        cell_map: array_t = None,
        cytoplasm_map: array_t = None,
        nucleus_map: array_t = None,
        safe_mode: bool = True,
    ) -> cell_t:
        """
        cell_map: Defined optional to follow cytoplasm and nucleus, but never None in fact
        Valid options: see CellAndCytoplasmMapsFromCombinations
        """
        if safe_mode:
            cell_map, cytoplasm_map, nucleus_map = CellAndCytoplasmMapsFromCombinations(
                cell=cell_map,
                cytoplasm=cytoplasm_map,
                nucleus=nucleus_map,
                should_return_nucleus=True,
            )
        cell = compartment_t.NewFromMap(cell_map)
        if cytoplasm_map is None:
            cytoplasm = None
        else:
            cytoplasm = cytoplasm_t.NewFromMap(cytoplasm_map, is_plain=False)
        if nucleus_map is None:
            nucleus = None
        else:
            nucleus = nucleus_t.NewFromMap(nucleus_map)

        instance = cls(
            label=label,
            nucleus=nucleus,
            cytoplasm=cytoplasm,
            centroid=cell.centroid,
            bb_slices=cell.bb_slices,
            touches_border=cell.touches_border,
            map_stream=cell.map_stream,
            area=cell.area,
        )

        return instance

    @classmethod
    def NewFromJsonString(
        cls,
        jsoned: str,
        /,
    ) -> cell_t:
        """"""
        return jsnr.ObjectFromJsonString(
            jsoned,
            builders={cls.__name__: cls.NewFromJsonDescription, "state_e": state_e},
        )

    def AddNucleus(self, nucleus_map: array_t, /) -> None:
        """
        For dividing cells
        """
        if self.nucleus is None:
            raise RuntimeError(f"Cannot add nucleus to a cell without a nucleus yet")

        nucleus = nucleus_t.NewFromMap(nucleus_map)
        self.nucleus = (self.nucleus, nucleus)

    @property
    def nuclei(self) -> Tuple[nucleus_t, ...]:
        """"""
        if self.nucleus is None:
            output = ()
        elif isinstance(self.nucleus, nucleus_t):
            output = (self.nucleus,)
        else:
            output = self.nucleus

        return output

    def Map(
        self,
        shape: Sequence[int],
        /,
        *,
        as_boolean: bool = False,
        with_labels: bool = False,
        margin: float = None,
    ) -> array_t:
        """
        with_labels: cytoplasm will be marked with label 1, nuclei with subsequent labels
        /!\ If with_labels, no margin should be requested. Indeed, what should be done with the nucleus map then?
        Apply the same margin or not? Or the method should take an additional margin parameter for the nucleus.
        """
        output = super().Map(shape, as_boolean=as_boolean, margin=margin)

        if with_labels:
            if as_boolean:
                raise ValueError('"with_labels" and "as_boolean" cannot be both True')

            if self.nucleus is not None:
                if isinstance(self.nucleus, nucleus_t):
                    nuclei = (self.nucleus,)
                else:
                    nuclei = self.nucleus
                for n_idx, nucleus in enumerate(nuclei, start=2):
                    map_ = nucleus.Map(shape, as_boolean=True)
                    output[map_] = n_idx

        return output

    def __str__(self) -> str:
        """"""
        super_str = super().__str__()

        lines = [super_str, "--- With compartments:"]
        initial_n_lines = lines.__len__()
        for compartment in (*self.nuclei, self.cytoplasm):
            if compartment is not None:
                lines.append(str(compartment))
        if lines.__len__() == initial_n_lines:
            return super_str

        return "\n".join(lines)


def CellAndCytoplasmMapsFromCombinations(
    *,
    cell: array_t = None,
    cytoplasm: array_t = None,
    nucleus: array_t = None,
    should_return_nucleus: bool = False,
) -> Union[Tuple[array_t, array_t], Tuple[array_t, array_t, array_t]]:
    """
    Valid options:
        - cell               => cytoplasm = None, nucleus = None
        - cell, cytoplasm    => nucleus = cell - cytoplasm
        - cell, nucleus      => cytoplasm = cell - nucleus
        - cytoplasm          => cell = filled cytoplasm, nucleus = cell - cytoplasm
        - cytoplasm, nucleus => cell = cytoplasm + nucleus
    """
    if cell is not None:
        cell = cell > 0
    if cytoplasm is not None:
        cytoplasm = cytoplasm > 0
    if nucleus is not None:
        nucleus = nucleus > 0
    filled_cytoplasm = None

    if (cell is None) and (cytoplasm is None):
        raise ValueError("Cytoplasm and cell arrays both None")
    if not ((cell is None) or (cytoplasm is None) or (nucleus is None)):
        raise ValueError("Nucleus, cytoplasm and cell arrays all not None")

    if not ((cell is None) or (cytoplasm is None)):
        filled_cytoplasm = ndimage_t.binary_fill_holes(cytoplasm)
        if not nmpy.array_equal(filled_cytoplasm, cell):
            raise ValueError("Cytoplasm outer borders do not coincide with cells")
    if not ((cell is None) or (nucleus is None)):
        if nmpy.any(cell[nucleus] == False):
            raise ValueError("Nuclei outer borders not restricted to cells")
    if not ((cytoplasm is None) or (nucleus is None)):
        if nmpy.any(cytoplasm[nucleus]):
            raise ValueError("Cytoplasm and nucleus arrays intersect")
        # Necessarily not already computed above since all 3 segmentations cannot be passed
        filled_cytoplasm = ndimage_t.binary_fill_holes(cytoplasm)
        union = nmpy.logical_or(cytoplasm, nucleus)
        if not nmpy.array_equal(filled_cytoplasm, union):
            raise ValueError("Cytoplasm inner borders do not coincide with nuclei")

    if cell is None:  # Then cytoplasm is not None
        if filled_cytoplasm is None:
            if nucleus is None:
                cell = ndimage_t.binary_fill_holes(cytoplasm)
            else:
                cell = cytoplasm.copy()
                cell[nucleus] = True
        else:
            cell = filled_cytoplasm
    # From then on, cell is not None

    if (cytoplasm is None) and (nucleus is not None):
        cytoplasm = cell.copy()
        cytoplasm[nucleus] = False
    if cytoplasm is not None:
        if filled_cytoplasm is None:
            filled_cytoplasm = ndimage_t.binary_fill_holes(cytoplasm)
        if nucleus is None:
            nucleus = filled_cytoplasm.copy()
            nucleus[cytoplasm] = False

        cytoplasms, n_cytoplasms = ndimage_t.label(filled_cytoplasm)  # connectivity=1
        nuclei, _ = ndimage_t.label(nucleus)  # connectivity=1
        for label in range(1, n_cytoplasms + 1):
            labels_inside = nmpy.unique(nuclei[cytoplasms == label])
            n_nuclei = labels_inside[labels_inside > 0].size
            if (n_nuclei < 1) or (n_nuclei > 2):
                raise ValueError(
                    f"{n_nuclei}: Invalid number of nuclei in cytoplasm. Expected=1 or 2."
                )

    if should_return_nucleus:
        if (nucleus is None) and (cytoplasm is not None):
            if filled_cytoplasm is None:
                filled_cytoplasm = ndimage_t.binary_fill_holes(cytoplasm)
            nucleus = filled_cytoplasm.copy()
            nucleus[cytoplasm] = False

        return cell, cytoplasm, nucleus

    return cell, cytoplasm



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import functools as fctl
from itertools import starmap, zip_longest
import json
from multiprocessing import Pool as pool_t
from operator import attrgetter as GetAttribute
from pathlib import Path as path_t
from typing import Any, Callable, Dict, Iterator, Optional, Sequence, Tuple, Union

import blosc as blsc
import numpy as nmpy
import scipy.ndimage as imge
from skimage.transform import resize as Resize

import cell_tracking_BC.in_out.file.jsoner as jsnr
import cell_tracking_BC.in_out.text.progress as prgs
import cell_tracking_BC.standard.issue as isse
from cell_tracking_BC.in_out.text.logger import LOGGER
from cell_tracking_BC.standard.number import MAX_INT
from cell_tracking_BC.standard.uid import Identity
from cell_tracking_BC.type.cell import cell_t
from cell_tracking_BC.type.cytoplasm import cytoplasm_t
from cell_tracking_BC.type.frame import frame_t, transform_h
from cell_tracking_BC.type.nucleus import nucleus_t
from cell_tracking_BC.type.segmentation import compartment_t, segmentation_t
from cell_tracking_BC.type.segmentations import segmentations_t
from cell_tracking_BC.type.track.forking import forking_track_t
from cell_tracking_BC.type.track.single import single_track_t
from cell_tracking_BC.type.track.structured import (
    feature_filtering_h,
    per_single_track_feature_h,
)
from cell_tracking_BC.type.track.unstructured import unstructured_track_t
from cell_tracking_BC.type.tracks.structured import tracks_t


array_t = nmpy.ndarray
#
all_versions_h = Dict[
    str, Tuple[Tuple[int, int], Union[Sequence[array_t], Sequence[frame_t]]]
]
channel_computation_h = Callable[[Dict[str, array_t], Dict[str, Any]], array_t]
morphological_feature_computation_h = Callable[
    [cell_t, Dict[str, Any]], Union[Any, Sequence[Any]]
]
radiometric_feature_computation_h = Callable[
    [cell_t, Union[array_t, Sequence[array_t]], Dict[str, Any]],
    Union[Any, Sequence[Any]],
]


@dtcl.dataclass(repr=False, eq=False)
class sequence_t(dict):
    """
    dict=Dict[str, Sequence[Union[bytes, array_t, frame_t]]]
    """

    path: Optional[path_t] = None
    shape: Tuple[int, int] = None
    original_length: int = None
    first_frame: int = None
    length: int = None
    base_channels: Sequence[str] = None
    # frames_of_channel: Dict[str, Sequence[Union[bytes, array_t, frame_t]]] = None
    # Name of channel whose frames store the segmented cells
    cell_channel: str = dtcl.field(init=False, default=None)
    segmentations: segmentations_t = dtcl.field(init=False, default=None)
    tracks: tracks_t = dtcl.field(init=False, default=None)

    @classmethod
    def NewFromFrames(
        cls,
        frames: array_t,
        in_channel_names: Sequence[Optional[str]],
        path: path_t,
        /,
        *,
        first_frame: int = 0,
        last_frame: int = MAX_INT,
        expected_shape: Sequence[int] = None,
    ) -> sequence_t:
        """
        in_channel_names: names equal to None or "___" or "---" indicate channels that should be discarded
        """
        # TODO: make this function accept various input shapes thanks to an additional arrangement parameter of the
        #     form THRC, T=time, H=channel, RC= row column. This requires that SequenceFromPath deals with TH combined
        #     dimension.
        if (n_dims := frames.ndim) not in (3, 4):
            raise ValueError(
                f"{n_dims}: Invalid number of dimensions of sequence with shape {frames.shape}. "
                f"Expected=3 or 4=(TIME POINTS*CHANNELS)xROWSxCOLUMNS or "
                f"TIME POINTSxCHANNELSxROWSxCOLUMNS."
            )
        n_in_channels = in_channel_names.__len__()

        if n_dims == 3:
            frames = frames[
                (first_frame * n_in_channels) : ((last_frame + 1) * n_in_channels), ...
            ]
        else:
            frames = frames[first_frame : (last_frame + 1), ...]
        first_frame = 0
        last_frame = MAX_INT

        if expected_shape is not None:
            if n_dims == 3:
                order_in = (0, 1, 2)
                order_out = (2, 0, 1)
                fixed_size = frames.shape[:1]
                to_be_resized = frames.shape[1:]
            else:
                order_in = (0, 1, 2, 3)
                order_out = (2, 3, 0, 1)
                fixed_size = frames.shape[:2]
                to_be_resized = frames.shape[2:]
            if to_be_resized != expected_shape:
                LOGGER.warn(
                    f"Resizing sequence from actual size {to_be_resized} (full shape={frames.shape}) "
                    f"to expected size {expected_shape}"
                )
                frames = nmpy.moveaxis(frames, order_in, order_out)
                frames = Resize(
                    frames, (*expected_shape, *fixed_size), preserve_range=True
                )
                frames = nmpy.moveaxis(frames, order_out, order_in)
                LOGGER.info(f"Resizing done. New sequence shape={frames.shape}")

        frames_of_channel = {}
        for name in in_channel_names:
            if (name is not None) and (name != "___") and (name != "---"):
                frames_of_channel[name] = []
        base_channel_names = tuple(frames_of_channel.keys())

        if n_dims == 3:
            c_idx = n_in_channels - 1
            time_point = -1
            for raw_frame in frames:
                c_idx += 1
                if c_idx == n_in_channels:
                    c_idx = 0
                    time_point += 1

                if time_point < first_frame:
                    continue
                elif time_point > last_frame:
                    break

                name = in_channel_names[c_idx]
                if name in base_channel_names:
                    frame = frame_t(raw_frame)
                    frames_of_channel[name].append(frame)
        else:
            for time_point, raw_frame in enumerate(frames):
                if time_point < first_frame:
                    continue
                elif time_point > last_frame:
                    break

                for c_idx, channel in enumerate(raw_frame):
                    name = in_channel_names[c_idx]
                    if name in base_channel_names:
                        frame = frame_t(channel)
                        frames_of_channel[name].append(frame)

        frames_of_base_channel = frames_of_channel[base_channel_names[0]]
        shape = frames_of_base_channel[0].shape
        length = frames_of_base_channel.__len__()
        instance = cls(
            path=path,
            shape=shape,
            original_length=frames.__len__(),
            first_frame=first_frame,
            length=length,
            base_channels=base_channel_names,
            # frames_of_channel=frames_of_channel,
        )
        instance.update(frames_of_channel)

        return instance

    @classmethod
    def NewFromJsonString(
        cls,
        jsoned: str,
        /,
    ) -> sequence_t:
        """"""
        return jsnr.ObjectFromJsonString(
            jsoned,
            builders={
                cls.__name__: cls.NewFromJsonDescription,
                "segmentations_t": segmentations_t.NewFromJsonString,
                "tracks_t": tracks_t.NewFromJsonString,
                "frame_t": frame_t.NewFromJsonString,
            },
        )

    @classmethod
    def NewFromJsonDescription(
        cls,
        description: Tuple[
            Dict[str, Sequence[Union[bytes, array_t, frame_t]]], Dict[str, Any]
        ],
        /,
    ) -> sequence_t:
        """"""
        frames, attributes = description

        instance = cls(
            path=attributes["path"],
            shape=attributes["shape"],
            original_length=attributes["original_length"],
            first_frame=attributes["first_frame"],
            length=attributes["length"],
            base_channels=attributes["base_channels"],
        )
        instance.update(frames)
        instance.cell_channel = attributes["cell_channel"]
        instance.segmentations = attributes["segmentations"]
        instance.tracks = attributes["tracks"]

        return instance

    def __len__(self) -> int:
        """"""
        return self.length

    @property
    def channels(self) -> Sequence[str]:
        """
        Names of channels read from file (base channels) and computed channels
        """
        # return tuple(self.frames_of_channel.keys())
        return tuple(self.keys())

    def Frames(
        self,
        /,
        *,
        channel: Union[str, Sequence[str]] = None,
    ) -> Union[
        Sequence[Union[array_t, frame_t]],
        Iterator[Sequence[Union[array_t, frame_t]]],
    ]:
        """
        channel: None=all (!) base channels; Otherwise, only (a) base channel name(s) can be passed
        as_iterator: Always considered True if channel is None or a sequence of channel names
        """
        if isinstance(channel, str):
            # return _AsArraySequence(self.frames_of_channel[channel])
            return _AsArraySequence(self[channel])
        else:
            return self._FramesForMultipleChannels(channel)

    def _FramesForMultipleChannels(
        self, channels: Optional[Sequence[str]], /
    ) -> Iterator[Sequence[Union[array_t, frame_t]]]:
        """
        /!\ If "channels" contains both base and non-base channels, then the returned tuples will contain both array_t
        and frame_t elements (but frame_t is a subclass of array_t, so...).
        """
        if channels is None:
            channels = self.base_channels

        for f_idx in range(self.length):
            # frames = tuple(self.frames_of_channel[_chl][f_idx] for _chl in channels)
            frames = tuple(self[_chl][f_idx] for _chl in channels)
            output = []
            for frame in frames:
                if isinstance(frame, bytes):
                    frame = blsc.unpack_array(frame)
                output.append(frame)

            yield output

    def ChannelExtrema(self, channel: str, /) -> Tuple[float, float]:
        """"""
        min_intensity = nmpy.Inf
        max_intensity = -nmpy.Inf

        # for frame in _AsArraySequence(self.frames_of_channel[channel]):
        for frame in _AsArraySequence(self[channel]):
            min_intensity = min(min_intensity, nmpy.amin(frame))
            max_intensity = max(max_intensity, nmpy.amax(frame))

        # Apparently, the intensities might not be Numpy object in certain situations (Numpy version?). Although
        # currently unexplained, an easy workaround is to return the intensities as is.
        try:
            output = min_intensity.item(), max_intensity.item()
        except AttributeError:
            output = min_intensity, max_intensity

        return output

    @property
    def has_cells(self) -> bool:
        """"""
        return self.cell_channel is not None

    def NCells(
        self, /, *, in_frame: Union[int, Sequence[int]] = None
    ) -> Union[int, Sequence[int]]:
        """
        in_frame: None=>total over the sequence
        """
        if in_frame is None:
            return sum(_cll.__len__() for _cll in self.cells_iterator)

        just_one = isinstance(in_frame, int)
        if self.has_cells:
            if just_one:
                in_frame = (in_frame,)

            output = in_frame.__len__() * [0]

            for f_idx, cells in enumerate(self.cells_iterator):
                if f_idx in in_frame:
                    output[in_frame.index(f_idx)] = cells.__len__()

            if just_one:
                output = output[0]
        elif just_one:
            output = 0
        else:
            output = in_frame.__len__() * (0,)

        return output

    @property
    def cell_frames(self) -> Sequence[frame_t]:
        """
        /!\ It is assumed that self.frames_of_channel[self.cell_channel] are frame_t's, not bytes
        """
        # return self.frames_of_channel[self.cell_channel]
        return self[self.cell_channel]

    @property
    def cells_iterator(self) -> Iterator[Sequence[cell_t]]:
        """"""
        for frame in self.cell_frames:
            yield frame.cells

    @property
    def cytoplasms_iterator(self) -> Iterator[Sequence[Optional[cytoplasm_t]]]:
        """"""
        for frame in self.cell_frames:
            yield tuple(map(GetAttribute("cytoplasm"), frame.cells))
            # output = []
            # for cell in frame.cells:
            #     output.append(cell.cytoplasm)
            # yield output

    @property
    def nuclei_iterator(self) -> Iterator[Sequence[Sequence[nucleus_t]]]:
        """"""
        for frame in self.cell_frames:
            yield tuple(map(GetAttribute("nuclei"), frame.cells))
            # output = []
            # for cell in frame.cells:
            #     output.append(cell.nuclei)
            # yield output

    def ApplyTransform(
        self,
        Transform: transform_h,
        /,
        *,
        channel: Union[str, Sequence[str]] = None,
        **kwargs,
    ) -> None:
        """
        /!\ It is assumed that self.frames_of_channel[channel] are frame_t's, not bytes
        channel: None=all (!)
        """
        if channel is None:
            channels = self.base_channels
        elif isinstance(channel, str):
            channels = (channel,)
        else:
            channels = channel

        for channel in channels:
            # targets = self.frames_of_channel[channel]
            targets = self[channel]
            if not isinstance(targets[0], frame_t):
                raise ValueError(
                    f"{type(targets[0]).__name__}: Invalid type of channel {channel} "
                    f"for {sequence_t.ApplyTransform.__name__}. Expected={frame_t.__name__}"
                )
            references_sets = self.Frames(channel=self.channels)
            for target, references in zip(targets, references_sets):
                refs_as_dict = dict(zip(self.channels, references))
                # refs_as_dict = {
                #     _nme: _frm for _nme, _frm in zip(self.channels, references)
                # }
                target.ApplyTransform(Transform, channels=refs_as_dict, **kwargs)

    def AddChannel(
        self, name: str, ChannelComputation: channel_computation_h, /, **kwargs
    ) -> None:
        """"""
        # if name in self.frames_of_channel:
        if name in self:
            raise ValueError(f"{name}: Existing channel cannot be overridden")

        computed = []
        for frames in self.Frames(channel=self.channels):
            frames_as_dict = dict(zip(self.channels, frames))
            # frames_as_dict = {_nme: _frm for _nme, _frm in zip(self.channels, frames)}
            frame = ChannelComputation(frames_as_dict, **kwargs)
            computed.append(frame)  # blsc.pack_array(frame))

        # self.frames_of_channel[name] = computed
        self[name] = computed

    def AddCellsFromSegmentations(
        self,
        channel: str,
        segmentations: segmentations_t,
    ) -> None:
        """
        Segmentation are supposed to be binary (as opposed to already labeled)
        """
        self.cell_channel = channel
        self.segmentations = segmentations

        for frame, segmentation in zip(self.cell_frames, segmentations):
            frame.AddCellsFromSegmentation(segmentation)

    def AddCellFeature(
        self,
        name: Union[str, Sequence[str]],
        Feature: Union[
            morphological_feature_computation_h, radiometric_feature_computation_h
        ],
        /,
        channel: Union[str, Sequence[str]] = None,
        should_run_in_parallel: bool = True,
        should_run_silently: bool = False,
        **kwargs,
    ) -> None:
        """
        name: If an str, then the value returned by Feature will be considered as a whole, whether it is actually a
        single value or a value container. If a sequence of str's, then the object returned by Feature will be iterated
        over, each element being matched with the corresponding name in "name".
        channel: if None, then morphological feature, else radiometric feature.

        /!\ There is no already-existing check
        """
        if isinstance(name, str):
            description = f'Feature "{name}"'
        elif name.__len__() > 2:
            description = f'Feature "{name[0]}, ..., {name[-1]}"'
        else:
            description = f'Feature "{name[0]}, {name[1]}"'
        ParallelFeature = fctl.partial(Feature, **kwargs)

        with prgs.ProgressDesign(should_be_silent=should_run_silently) as progress:
            if channel is None:
                iterator = self.cells_iterator
            else:
                # iterator = zip(self.cells_iterator, _AsArraySequence(self.frames_of_channel[channel]))
                iterator = zip(self.cells_iterator, _AsArraySequence(self[channel]))
            progress_context = prgs.progress_context_t(
                progress,
                iterator,
                total=self.length,
                description=description,
            )

            if should_run_in_parallel:
                pool = pool_t()
                MapFunctionOnList = pool.map
                StarMapFunctionOnList = pool.starmap
            else:
                pool = None
                MapFunctionOnList = map
                StarMapFunctionOnList = starmap

            if channel is None:
                if isinstance(name, str):
                    for cells in progress_context.elements:
                        features = MapFunctionOnList(ParallelFeature, cells)
                        for cell, feature in zip(cells, features):
                            cell.AddFeature(name, feature)
                else:
                    names = name
                    for cells in progress_context.elements:
                        multi_features = MapFunctionOnList(ParallelFeature, cells)
                        for cell, features in zip(cells, multi_features):
                            for name, feature in zip(names, features):
                                cell.AddFeature(name, feature)
            else:
                if isinstance(name, str):
                    for cells, frame in progress_context.elements:
                        features = StarMapFunctionOnList(
                            ParallelFeature,
                            zip_longest(cells, (frame,), fillvalue=frame),
                        )
                        for cell, feature in zip(cells, features):
                            cell.AddFeature(name, feature)
                else:
                    names = name
                    for cells, frame in progress_context.elements:
                        multi_features = StarMapFunctionOnList(
                            ParallelFeature,
                            zip_longest(cells, (frame,), fillvalue=frame),
                        )
                        for cell, features in zip(cells, multi_features):
                            for name, feature in zip(names, features):
                                cell.AddFeature(name, feature)

            if should_run_in_parallel:
                pool.close()
                pool.terminate()

    @property
    def available_cell_features(self) -> Sequence[str]:
        """"""
        one_cell = self.cell_frames[0].cells[0]

        return one_cell.available_features

    def AddTracks(
        self,
        tracks: tracks_t,
        /,
    ) -> None:
        """"""
        self.tracks = tracks

    def AddTrackFeature(
        self,
        name: str,
        cell_feature_name: Union[str, Sequence[str]],
        FeatureComputation: feature_filtering_h,
        /,
        *args,
        geometric_mode: bool = False,
        **kwargs,
    ) -> None:
        """
        /!\ There is no already-existing check
        """
        if isinstance(cell_feature_name, str):
            cell_feature_names = (cell_feature_name,)
        else:
            cell_feature_names = cell_feature_name
        if any(_ftr not in self.available_cell_features for _ftr in cell_feature_names):
            raise ValueError(f"{cell_feature_name}: Invalid cell feature(s)")

        for track in self.tracks:
            track.AddFeature(
                name, cell_feature_names, FeatureComputation, *args, geometric_mode=geometric_mode, **kwargs
            )

    def CellFeature(
        self, feature: str, /, *, geometric_mode: bool = False
    ) -> Dict[int, Tuple[single_track_t, Sequence[Any]]]:
        """"""
        output = {}

        if geometric_mode:
            tracks = self.tracks.geometric_single_tracks_iterator
        else:
            tracks = self.tracks.single_tracks_iterator
        for track in tracks:
            output[track.label] = (
                track,
                track.CellFeature(feature, geometric_mode=geometric_mode),
            )

        return output

    def TrackFeature(self, feature: str, /) -> per_single_track_feature_h:
        """"""
        output = {}

        for track in self.tracks:
            output.update(track.features[feature])

        return output

    def DescriptionOfFrame(self, frame_idx: int, /) -> Dict[str, Any]:
        """"""
        output = {}

        track_labels = []
        for cell in self.cell_frames[frame_idx].cells:
            labels = self.tracks.TrackLabelsContainingCell(cell, tolerant_mode=True)
            if labels is None:
                text = "i"  # Invalid
            elif labels.__len__() == 0:
                text = "p"  # Pruned
            elif labels.__len__() > 1:
                text = "\n".join(str(_lbl) for _lbl in labels)
            else:
                text = str(labels[0])
            track_labels.append((text, cell.centroid))
        output["track_labels"] = tuple(track_labels)

        return output

    def ClearContents(self):
        """
        To free up some memory when all processing has been done
        """
        for frames in self.Frames():
            for frame in frames:
                frame.ClearContents()

    def PrintValidInvalidTrackSummary(self) -> None:
        """"""
        n_invalids = []
        issues_per_type = {}
        for track_type in (unstructured_track_t, single_track_t, forking_track_t):
            invalids = [
                _rcd for _rcd in self.tracks.invalids if isinstance(_rcd[0], track_type)
            ]
            track_type = track_type.__name__[:-2].replace("_", " ").capitalize()
            if invalids.__len__() == 0:
                number = f"{track_type}: None"
            else:
                number = f"{track_type}: {invalids.__len__()}"

                issues = (
                    f"    Track {_tck.labels_as_str}{isse.ISSUE_SEPARATOR}"
                    f"{isse.FactorizedIssuesAsStr(_iss, max_n_contexts=5)}"
                    for _tck, _iss in invalids
                )
                issues_per_type[track_type] = "\n".join(issues)

            n_invalids.append(number)

        n_invalids = ", ".join(n_invalids)
        issues_per_type = "\n".join(
            f"{_typ}:\n{_iss}" for _typ, _iss in issues_per_type.items()
        )

        print(
            f"Tracks: valid={self.tracks.__len__()}; "
            f"invalid={n_invalids}; "
            f"fully pruned={self.tracks.fully_pruned.__len__()}\n"
            f"{issues_per_type}"
        )

    def AsJsonString(self) -> str:
        """"""
        # Do not use dtcl.asdict(self) since it recurses into dataclass instances which, if they extend a "container"
        # class like list or dict, will lose the contents.
        attributes = {_fld.name: getattr(self, _fld.name) for _fld in dtcl.fields(self)}

        # dict(self): To hide this method, which would otherwise be infinitely recursively called
        return jsnr.JsonStringOf((dict(self), attributes), true_type=self)

    def __repr__(self) -> str:
        """"""
        return Identity(self)

    def __str__(self) -> str:
        """"""
        if self.tracks is None:
            invalid_tracks = None
            fully_pruned_tracks = None
        else:
            invalid_tracks = str(self.tracks.invalids)
            fully_pruned_tracks = str(self.tracks.fully_pruned)

        all_extrema = []
        for channel in self.channels:
            extrema = self.ChannelExtrema(channel)
            all_extrema.append(f"{channel}=[{extrema[0]},{extrema[1]}]")
        all_extrema = "\n    ".join(all_extrema)

        return (
            f"{repr(self)}:\n"
            f"    {self.path=}\n"
            f"    {self.base_channels=}\n"
            f"    {self.channels=}\n"
            f"    {all_extrema}\n"
            f"    {self.shape=}\n"
            f"    {self.original_length=}\n"
            f"    {self.length=}\n"
            f"    {self.tracks=}\n"
            f"    {invalid_tracks=}\n"
            f"    {fully_pruned_tracks=}\n"
        )

    def __eq__(self, other) -> bool:
        """"""
        if hasattr(other, "AsJsonString"):
            return self.AsJsonString() == other.AsJsonString()

        return False

    def __hash__(self) -> int:
        """"""
        return hash((self.path, self.shape, self.first_frame, self.length))


sequence_h = Union[
    Sequence[array_t], Sequence[segmentation_t], segmentations_t, sequence_t
]


def _AsArraySequence(
    frames: Union[Sequence[bytes], Sequence[array_t]], /
) -> Sequence[array_t]:
    """"""
    if isinstance(frames[0], bytes):
        return tuple(map(blsc.unpack_array, frames))

    return frames


def _JsonableFrame(
    frame: Union[bytes, array_t, frame_t], /
) -> Tuple[str, Union[array_t, str]]:
    """"""
    if isinstance(frame, bytes):
        output = ("bytes", frame)  # pcst.BStream2PCA(frame))
    # Test frame_t first (instead of array_t) since a frame_t is also an array_t
    elif isinstance(frame, frame_t):
        output = ("frame_t", frame.AsJsonString())
    else:
        output = ("array_t", frame)

    return output  # jsnr.JsonStringOf(output)


def _FrameFromJsonString(jsoned: Sequence, /) -> Union[bytes, array_t, frame_t]:
    """"""
    if isinstance(jsoned, str):
        jsoned = json.loads(jsoned)
    type_, frame = jsoned
    if type_ not in ("array_t", "bytes", "frame_t"):
        type_ = json.loads(type_)

    if type_ == "bytes":
        return bytes.fromhex(
            json.loads(frame)
        )  # .encode(encoding="raw-unicode-escape")#pcst.PCA2BStream(nmpy.array(json.dumps(frame)))
    if type_ == "array_t":
        return nmpy.array(json.loads(frame))

    return frame_t.NewFromJsonString(frame)


def BoundingBoxSlices(sequence: array_t, /) -> Tuple[slice, slice]:
    """
    sequence: as an XYCT-volume
    """
    min_reduction = nmpy.amin(sequence, axis=-1)
    max_reduction = nmpy.amax(sequence, axis=-1)
    variable = nmpy.any(min_reduction != max_reduction, axis=-1)

    output = imge.find_objects(variable.astype(nmpy.int8))
    if output.__len__() == 0:
        raise ValueError(f"{sequence}: Empty sequence")

    return output[0]


def AllChannelsOfSequence(
    sequence: Union[Sequence[array_t], sequence_t]
) -> Tuple[all_versions_h, str]:
    """"""
    if isinstance(sequence, sequence_t):
        all_channels = {}
        for channel in sequence.channels:
            frames = sequence.Frames(channel=channel)
            min_value = min(nmpy.amin(_frm) for _frm in frames)
            max_value = max(nmpy.amax(_frm) for _frm in frames)
            all_channels[channel] = ((min_value, max_value), frames)
        current_channel = sequence.channels[0]
    else:
        current_channel = "MAIN"
        min_value = min(nmpy.amin(_frm) for _frm in sequence)
        max_value = max(nmpy.amax(_frm) for _frm in sequence)
        all_channels = {current_channel: ((min_value, max_value), sequence)}

    return all_channels, current_channel


def AllSegmentationsOfSequence(sequence: sequence_h) -> Tuple[all_versions_h, str]:
    """"""
    if isinstance(sequence, (segmentations_t, sequence_t)):
        if isinstance(sequence, segmentations_t):
            segmentations = sequence
        else:
            segmentations = sequence.segmentations

        all_versions = {}
        compartments, versions = segmentations.available_versions
        for compartment in compartments:
            for version in versions:
                key = f"{compartment.name}:{version[0]}:{version[1]}"
                frames = segmentations.CompartmentsWithVersion(
                    compartment, index=version[0], name=version[1]
                )
                all_versions[key] = ((0, 1), frames)
        current_version = f"{compartment_t.CELL.name}:{versions[0][0]}:{versions[0][1]}"
    else:
        current_version = "MAIN"
        all_versions = {current_version: ((0, 1), sequence)}

    return all_versions, current_version


def AllStreamsOfSequence(sequence: sequence_t) -> Tuple[all_versions_h, str]:
    """"""
    all_streams, current_stream = AllChannelsOfSequence(sequence)
    all_versions, _ = AllSegmentationsOfSequence(sequence)

    all_streams.update(all_versions)

    return all_streams, current_stream



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from enum import Enum as enum_t
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import numpy as nmpy
import pca_b_stream as pcst
import scipy.ndimage as spim
import scipy.optimize as spop
import skimage.morphology as mrph
import skimage.segmentation as sgmt
from scipy.spatial import distance as dstc

import cell_tracking_BC.in_out.file.jsoner as jsnr
import cell_tracking_BC.task.registration as rgst
import cell_tracking_BC.type.cell as cll_
from cell_tracking_BC.in_out.text.logger import LOGGER
from cell_tracking_BC.task.jaccard import PseudoJaccard


array_t = nmpy.ndarray


class compartment_t(enum_t):
    CELL = 0
    CYTOPLASM = 1
    NUCLEUS = 2


_CELL_IDX = compartment_t.CELL.value
_CYTO_IDX = compartment_t.CYTOPLASM.value


version_id_h = Tuple[int, str]
versioned_compartment_h = Union[bytes, array_t, None, version_id_h]
version_h = Tuple[versioned_compartment_h, versioned_compartment_h]
versions_h = Dict[version_id_h, version_h]

CellIssues_h = Callable[[int, array_t, dict], Optional[Union[str, Sequence[str]]]]


@dtcl.dataclass(repr=False, eq=False)
class segmentation_t:

    cell: Union[bytes, array_t] = None  # dtype=bool; Never None after instantiation
    cytoplasm: Union[bytes, array_t] = None  # dtype=bool
    # nucleus = cell - cytoplasm
    version_idx: int = 0
    version_name: str = "original"
    # version: See property "version"
    versions: versions_h = dtcl.field(init=False, default_factory=dict)
    invalid_cells: List[Tuple[int, Union[str, Sequence[str]]]] = dtcl.field(
        init=False, default_factory=list
    )

    def __post_init__(self) -> None:
        """"""
        version = self._NewVersionFromCompartments(self.cell, self.cytoplasm)
        self.versions[self.version] = version

    @classmethod
    def NewFromCompartments(
        cls,
        /,
        *,
        cell: array_t = None,
        cytoplasm: array_t = None,
        nucleus: array_t = None,
    ) -> segmentation_t:
        """
        Valid options: see cell.CellAndCytoplasmMapsFromCombinations
        """
        cell, cytoplasm = cll_.CellAndCytoplasmMapsFromCombinations(
            cell=cell, cytoplasm=cytoplasm, nucleus=nucleus
        )

        cell = _AsMostConcise(cell)
        if cytoplasm is not None:
            cytoplasm = _AsMostConcise(cytoplasm)

        return cls(cell=cell, cytoplasm=cytoplasm)

    @classmethod
    def NewFromDict(cls, dictionary: Dict[str, Any], /) -> segmentation_t:
        """"""
        instance = cls()

        if isinstance(cell := dictionary["cell"], bytes):
            instance.cell = cell
        else:
            instance.cell = nmpy.array(cell)
        if isinstance(cytoplasm := dictionary["cytoplasm"], bytes):
            instance.cytoplasm = cytoplasm
        elif cytoplasm is not None:
            instance.cytoplasm = nmpy.array(cytoplasm)
        instance.version_idx = dictionary["version_idx"]
        instance.version_name = dictionary["version_name"]
        instance.versions = {}

        versions = instance.versions
        for version, values in dictionary["versions"].items():
            new_values = []
            for v_idx, value in enumerate(values):
                if (value is None) or isinstance(value, bytes) or IsVersion(value):
                    new_value = value
                else:
                    new_value = nmpy.array(value)
                new_values.append(new_value)
            versions[version] = tuple(new_values)

        return instance

    @classmethod
    def NewFromJsonString(cls, jsoned: str, /) -> segmentation_t:
        """"""
        return jsnr.ObjectFromJsonString(
            jsoned, builders={cls.__name__: cls.NewFromJsonDescription}
        )

    @classmethod
    def NewFromJsonDescription(cls, description: dict) -> segmentation_t:
        #
        instance = cls.__new__(cls)
        for key, value in description.items():
            setattr(instance, key, value)

        return instance

    def AddFakeVersion(self, version: str, /) -> None:
        """"""
        self._AddVersionForCompartments(
            self.cell, self.cytoplasm, version, (False, False)
        )

    def _AddVersionForCompartments(
        self,
        cell: Union[bytes, array_t],
        cytoplasm: Optional[Union[bytes, array_t]],
        version: str,
        has_changed: Tuple[bool, bool],
        /,
    ) -> None:
        """
        cell and cytoplasm are supposed to already be in their most concise format
        """
        if has_changed[0]:
            self.cell = cell
        else:
            cell = self.version
        if has_changed[1]:
            self.cytoplasm = cytoplasm
        else:
            cytoplasm = self.version
        new_version = self._NewVersionFromCompartments(cell, cytoplasm)

        self.version_idx += 1
        self.version_name = version
        self.versions[self.version] = new_version

    def _NewVersionFromCompartments(
        self,
        cell: Union[bytes, array_t, version_id_h],
        cytoplasm: Optional[Union[bytes, array_t, version_id_h]],
        /,
    ) -> version_h:
        """"""
        output = [None, None]

        if IsVersion(cell):
            cell = self._ActualOriginal(cell, _CELL_IDX)
        if IsVersion(cytoplasm):
            cytoplasm = self._ActualOriginal(cytoplasm, _CYTO_IDX)

        output[_CELL_IDX] = cell
        output[_CYTO_IDX] = cytoplasm

        return output[0], output[1]

    def _ActualOriginal(self, version: version_id_h, index: int, /) -> version_id_h:
        """"""
        output = version

        while IsVersion(self.versions[output][index]):
            output = self.versions[output][index]

        return output

    @property
    def nucleus(self) -> array_t:
        """"""
        if self.cytoplasm is None:
            raise RuntimeError(
                "Requesting nucleus map in a segmentation w/o cytoplasm map"
            )

        return nmpy.logical_xor(AsArray(self.cell), AsArray(self.cytoplasm))

    def NonNoneAsList(self) -> Tuple[List[array_t], Sequence[str]]:
        """
        For preparing calls to cell_tracking_BC.type.cell_t.NewFromMaps
        """
        if self.cytoplasm is None:
            segmentations = [AsArray(self.cell)]
            parameters = ("cell_map",)
        else:
            segmentations = [AsArray(self.cytoplasm), AsArray(self.cell)]
            parameters = ("cytoplasm_map", "cell_map")

        return segmentations, parameters

    @property
    def version(self) -> version_id_h:
        """"""
        return self.version_idx, self.version_name

    @property
    def available_versions(
        self,
    ) -> Tuple[Tuple[compartment_t, ...], Sequence[version_id_h]]:
        """
        The sequences contain tuples (version index, version basename, version name), where version name is a
        combination of version index and basename.
        """
        if self.cytoplasm is None:
            compartments = (compartment_t.CELL,)
        else:
            compartments = (
                compartment_t.CELL,
                compartment_t.CYTOPLASM,
                compartment_t.NUCLEUS,
            )

        versions = []
        for version in self.versions.keys():
            versions.append(version)

        return compartments, versions

    def LatestCompartment(self, compartment: compartment_t, /) -> array_t:
        """"""
        if compartment is compartment_t.CELL:
            return AsArray(self.cell)

        if compartment is compartment_t.CYTOPLASM:
            if self.cytoplasm is not None:
                return AsArray(self.cytoplasm)
        elif compartment is compartment_t.NUCLEUS:
            return self.nucleus

        raise ValueError(f"{compartment}: Invalid compartment, or compartment is None")

    def CompartmentWithVersion(
        self, compartment: compartment_t, /, *, index: int = None, name: str = None
    ) -> array_t:
        """
        index: if None and name is None also, then returns latest version
        """
        if (
            ((index is None) and (name is None))
            or (index == self.version_idx)
            or (name == self.version_name)
        ):
            return self.LatestCompartment(compartment)

        if index is None:
            VersionIsAMatch = lambda _ver: name == _ver[1]
        elif name is None:
            VersionIsAMatch = lambda _ver: index == _ver[0]
        else:
            VersionIsAMatch = lambda _ver: (index, name) == _ver
        for version in self.versions.keys():
            if VersionIsAMatch(version):
                output = self._ResolvedCompartmentWithVersion(compartment, version)
                if output is None:
                    break
                else:
                    return output

        raise ValueError(
            f"{compartment}/{index}/{name}: Invalid compartment/index/name combination"
        )

    def _ResolvedCompartmentWithVersion(
        self, compartment: compartment_t, version: version_id_h, /
    ) -> Optional[array_t]:
        """"""
        if compartment is compartment_t.NUCLEUS:
            if self.cytoplasm is None:
                output = None
            else:
                cell = self.versions[version][_CELL_IDX]
                cytoplasm = self.versions[version][_CYTO_IDX]
                if IsVersion(cell):
                    cell = self.versions[cell][_CELL_IDX]
                if IsVersion(cytoplasm):
                    cytoplasm = self.versions[cytoplasm][_CYTO_IDX]
                output = nmpy.logical_xor(AsArray(cell), AsArray(cytoplasm))
        else:
            output = self.versions[version][compartment.value]
            if IsVersion(output):
                output = self.versions[output][compartment.value]
            output = AsArray(output)

        return output

    @property
    def cell_areas(self) -> Sequence[int]:
        """"""
        labeled, n_cells = mrph.label(
            AsArray(self.cell), return_num=True, connectivity=1
        )
        output = tuple(
            nmpy.count_nonzero(labeled == _lbl) for _lbl in range(1, n_cells + 1)
        )

        return output

    def CellAreaHistogram(
        self,
        /,
        *,
        n_bins: int = None,
        should_return_centers: bool = False,
        should_round_centers: bool = True,
    ) -> Tuple[array_t, array_t]:
        """"""
        areas = self.cell_areas

        if n_bins is None:
            n_bins = int(round(nmpy.sqrt(areas.__len__())))
        counts, edges = nmpy.histogram(areas, bins=n_bins)
        if should_return_centers:
            centers = 0.5 * (edges[:-1] + edges[1:])
            if should_round_centers:
                centers = nmpy.around(centers).astype(nmpy.uint64)
            edges = centers

        return counts, edges

    def ClearBorderObjects(self) -> None:
        """"""
        if self.cytoplasm is None:
            originals = (AsArray(self.cell),)
        else:
            originals = (AsArray(self.cytoplasm), AsArray(self.cell))
        new_version = [_cpt.copy() for _cpt in originals]

        # --- Clear outer segmentation border objects
        has_changed = _ClearBorderObjects(new_version, originals, -1)
        if more_than_1 := (new_version.__len__() > 1):
            # --- Clear inner segmentation border objects
            has_changed |= _ClearBorderObjects(new_version, originals, 0)

        cell = _AsMostConcise(new_version[-1])
        if more_than_1:
            cytoplasm = _AsMostConcise(new_version[0])
            has_changed = (has_changed, has_changed)
        else:
            cytoplasm = None
            has_changed = (has_changed, False)
        self._AddVersionForCompartments(cell, cytoplasm, "cleared borders", has_changed)

    def FilterCellsOut(
        self,
        CellIssues: CellIssues_h,
        /,
        **kwargs,
    ) -> None:
        """
        Currently, only applicable to the cell segmentation when no other compartments are present

        Parameters
        ----------
        CellIssues: Arguments are: cell label (from 1), labeled segmentation, and (optional) keyword arguments; Returned
            value can be None, an str, or a sequence of str.
        kwargs: Passed to CellIsInvalid as keyword arguments
        """
        assert self.cytoplasm is None

        compartment = AsArray(self.cell, as_copy=True)

        labeled, n_cells = mrph.label(compartment, return_num=True, connectivity=1)
        invalids = []
        for label in range(1, n_cells + 1):
            issues = CellIssues(label, labeled, **kwargs)
            if issues is not None:
                invalids.append((label, issues))

        if has_changed := (invalids.__len__() > 0):
            for invalid in invalids:
                compartment[labeled == invalid[0]] = 0
                self.invalid_cells.append(invalid)

        self._AddVersionForCompartments(
            _AsMostConcise(compartment),
            self.cytoplasm,
            "filtered cells",
            (has_changed, False),
        )

    def CorrectBasedOnTemporalCoherence(
        self,
        previous: segmentation_t,
        /,
        *,
        min_jaccard: float = 0.75,
        max_area_discrepancy: float = 0.25,
        min_cell_area: int = 0,
        time_point: int = None,
    ) -> int:
        """
        min_jaccard: Actually, Pseudo-Jaccard

        Currently, only applicable to the cell segmentation when no other compartments are present
        """
        assert self.cytoplasm is None

        current = AsArray(self.cell, as_copy=True)
        previous = AsArray(previous.cell)
        if time_point is None:
            at_time_point = ""
        else:
            at_time_point = f" at time point {time_point}"

        labeled_current, n_cells_current = mrph.label(
            current, return_num=True, connectivity=1
        )
        labeled_previous, n_cells_previous = mrph.label(
            previous, return_num=True, connectivity=1
        )
        jaccards = _PairwiseJaccards(
            n_cells_previous, n_cells_current, labeled_previous, labeled_current
        )
        c_to_p_links = _CurrentToPreviousLinks(jaccards, min_jaccard)

        n_corrections = 0
        for label_current in range(1, n_cells_current + 1):
            labels_previous = c_to_p_links.get(label_current - 1)
            if (labels_previous is not None) and (labels_previous.__len__() > 1):
                labels_previous = nmpy.array(labels_previous) + 1
                context = f"between cells {sorted(labels_previous)} and fused cell {label_current}{at_time_point}"

                where_fused = labeled_current == label_current

                split = _SplitLabels(
                    where_fused,
                    labeled_previous,
                    labels_previous,
                    max_area_discrepancy,
                    min_cell_area,
                    context,
                )
                if split is None:
                    continue

                current[where_fused] = 0
                current[split > 0] = 1
                n_corrections += 1

        self._AddVersionForCompartments(
            _AsMostConcise(current),
            self.cytoplasm,
            "corrected w/ temp corr",
            (n_corrections > 0, False),
        )

        return n_corrections

    def AsDict(self) -> Dict[str, Any]:
        """"""
        if isinstance(self.cell, bytes):
            cell = self.cell
        else:
            cell = self.cell.tolist()
        if self.cytoplasm is None:
            cytoplasm = None
        elif isinstance(self.cytoplasm, bytes):
            cytoplasm = self.cytoplasm
        else:
            cytoplasm = self.cytoplasm.tolist()
        output = {
            "cell": cell,
            "cytoplasm": cytoplasm,
            "version_idx": self.version_idx,
            "version_name": self.version_name,
            "versions": {},
        }

        versions = output["versions"]
        for name, values in self.versions.items():
            new_values = []
            for value in values:
                if (value is None) or isinstance(value, bytes) or IsVersion(value):
                    new_value = value
                else:
                    new_value = value.tolist()
                new_values.append(new_value)
            versions[name] = new_values

        return output

    def AsJsonString(self) -> str:
        """"""
        return jsnr.JsonStringOf(self)

    def __eq__(self, other) -> bool:
        """"""
        if hasattr(other, "AsJsonString"):
            return self.AsJsonString() == other.AsJsonString()

        return False


def IsVersion(maybe: Any, /) -> bool:
    """"""
    return (
        isinstance(maybe, tuple)
        and (maybe.__len__() == 2)
        and isinstance(maybe[0], int)
        and isinstance(maybe[1], str)
    )


def _AsMostConcise(compartment: array_t, /) -> Union[bytes, array_t]:
    """"""
    return compartment
    # map_stream = pcst.PCA2BStream(compartment.astype(nmpy.bool_, copy=False))
    # if map_stream.__len__() < compartment.nbytes:
    #     return map_stream
    # else:
    #     return compartment


def AsArray(compartment: Union[bytes, array_t], /, *, as_copy: bool = False) -> array_t:
    """"""
    if isinstance(compartment, bytes):
        return pcst.BStream2PCA(compartment)
    elif as_copy:
        return nmpy.array(compartment)
    else:
        return compartment


def _ClearBorderObjects(
    compartments: List[array_t], originals: Sequence[array_t], idx: int, /
) -> bool:
    """"""
    compartment = compartments[idx]
    sgmt.clear_border(compartment, in_place=True)

    return not nmpy.array_equal(compartment, originals[idx])


def _PairwiseJaccards(
    n_cells_previous: int,
    n_cells_current: int,
    labeled_previous: array_t,
    labeled_current: array_t,
    /,
) -> array_t:
    """"""
    labels_previous = nmpy.fromiter(range(1, n_cells_previous + 1), dtype=nmpy.uint64)
    labels_current = nmpy.fromiter(range(1, n_cells_current + 1), dtype=nmpy.uint64)
    # Note the reversed parameter order in PseudoJaccard since a fusion is a division in reversed time
    _PseudoJaccard = lambda lbl_1, lbl_2: PseudoJaccard(
        labeled_current, labeled_previous, lbl_2, lbl_1
    )

    output = dstc.cdist(
        labels_previous[:, None], labels_current[:, None], metric=_PseudoJaccard
    )

    return output


def _CurrentToPreviousLinks(
    pairwise_jaccards: array_t, min_jaccard: float, /
) -> Dict[int, Sequence[int]]:
    """"""
    output = {}

    while True:
        row_idc, col_idc = spop.linear_sum_assignment(1.0 - pairwise_jaccards)
        valid_idc = pairwise_jaccards[row_idc, col_idc] > min_jaccard
        if not nmpy.any(valid_idc):
            break

        valid_row_idc = row_idc[valid_idc]
        for key, value in zip(col_idc[valid_idc], valid_row_idc):
            if key in output:
                output[key].append(value)
            else:
                output[key] = [value]

        pairwise_jaccards[valid_row_idc, :] = 0.0

    return output


def _SplitLabels(
    where_fused: array_t,
    labeled_previous: array_t,
    labels_previous: array_t,
    max_area_discrepancy: float,
    min_cell_area: int,
    context: str,
    /,
) -> Optional[array_t]:
    """"""
    output = nmpy.zeros_like(labeled_previous)
    for l_idx, previous_label in enumerate(labels_previous, start=1):
        output[labeled_previous == previous_label] = l_idx

    fused_area = nmpy.count_nonzero(where_fused)
    split_area = nmpy.count_nonzero(output)
    discrepancy = abs(split_area - fused_area) / fused_area
    if discrepancy > max_area_discrepancy:
        LOGGER.warn(
            f"Segmentation correction discarded due to high t-total-area/(t+1)-fused-area discrepancy "
            f"{context}: "
            f"Actual={discrepancy}. Expected<={max_area_discrepancy}."
        )
        return None

    (fused_local, split_local), corners = rgst.InCommonNonZeroRectangles(
        where_fused, output, for_rotation=True
    )

    angle = rgst.RotationInBinary(fused_local, split_local > 0)
    rotated = rgst.RotatedLabeled(split_local, angle)

    split_local = _DilatedWithLabelPreservation(rotated, fused_local, min_cell_area)
    if split_local is None:
        LOGGER.warn(
            f"Segmentation correction discarded due to invalid split regions "
            f"{context}"
        )
        return None

    corner = corners[0]
    rows_gbl, cols_gbl = nmpy.meshgrid(
        range(corner[0], corner[0] + split_local.shape[0]),
        range(corner[1], corner[1] + split_local.shape[1]),
        indexing="ij",
    )
    rows_lcl, cols_lcl = nmpy.indices(split_local.shape)
    valid = nmpy.logical_and(
        nmpy.logical_and(rows_gbl >= 0, cols_gbl >= 0),
        nmpy.logical_and(rows_gbl < output.shape[0], cols_gbl < output.shape[1]),
    )
    output.fill(0)
    output[rows_gbl[valid], cols_gbl[valid]] = split_local[
        rows_lcl[valid], cols_lcl[valid]
    ]

    return output


def _DilatedWithLabelPreservation(
    labeled: array_t, roi: array_t, min_cell_area: int, /
) -> Optional[array_t]:
    """"""
    output = nmpy.ones_like(labeled)

    n_labels = nmpy.amax(labeled).item()
    non_roi = nmpy.logical_not(roi)

    distance_map = spim.distance_transform_edt(labeled != 1)
    for label in range(2, n_labels + 1):
        current_map = spim.distance_transform_edt(labeled != label)
        closer_bmap = current_map < distance_map
        output[closer_bmap] = label
        distance_map[closer_bmap] = current_map[closer_bmap]

    while True:
        intermediate = nmpy.zeros_like(output)
        for label in range(1, n_labels + 1):
            where = output == label
            if not nmpy.any(where):
                return None
            # TODO: check how erosion behaves on image borders
            eroded = mrph.binary_erosion(where)
            intermediate[eroded] = label
        nmpy.copyto(output, intermediate)
        output[non_roi] = 0

        eroded_labeled, n_eroded_labels = mrph.label(
            output > 0, return_num=True, connectivity=1
        )
        areas = tuple(
            nmpy.count_nonzero(eroded_labeled == _lbl)
            for _lbl in range(1, n_eroded_labels + 1)
        )
        for label, area in enumerate(areas, start=1):
            if area < min_cell_area:
                output[eroded_labeled == label] = 0
                n_eroded_labels -= 1
        if n_eroded_labels == n_labels:
            break

    return output
